# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from typing import Dict, List, Set, Callable
from dlk.core.base_module import SimpleModule, BaseModuleConfig
from . import decoder_register, decoder_config_register
from dlk.core.modules import module_config_register, module_register

@decoder_config_register("linear")
class LinearConfig(BaseModuleConfig):
    """Config for Linear 

    Config Example:
        >>> {
        >>>     "module": {
        >>>         "_base": "linear",
        >>>     },
        >>>     "config": {
        >>>         "input_size": "*@*",
        >>>         "output_size": "*@*",
        >>>         "pool": null,
        >>>         "dropout": 0.0,
        >>>         "output_map": {},
        >>>         "input_map": {}, // required_key: provide_key
        >>>     },
        >>>     "_link":{
        >>>         "config.input_size": ["module.config.input_size"],
        >>>         "config.output_size": ["module.config.output_size"],
        >>>         "config.pool": ["module.config.pool"],
        >>>         "config.dropout": ["module.config.dropout"],
        >>>     },
        >>>     "_name": "linear",
        >>> }
    """
    def __init__(self, config: Dict):
        super(LinearConfig, self).__init__(config)
        self.linear_config = config["module"]
        self.post_check(config['config'], used=[
            "input_size",
            "output_size",
            "pool",
            "dropout",
            "return_logits",
        ])


@decoder_register("linear")
class Linear(SimpleModule):
    """wrap for torch.nn.Linear
    """
    def __init__(self, config: LinearConfig):
        super(Linear, self).__init__(config)
        self._provide_keys = {'logits'}
        self._required_keys = {'embedding'}
        self._provided_keys = set()

        self.config = config

        self.linear = module_register.get('linear')(module_config_register.get('linear')(config.linear_config))

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns: 
            None

        """
        self.linear.init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """All step do this

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        inputs[self.get_output_name("logits")] = self.linear(inputs[self.get_input_name('embedding')])
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([inputs[self.get_output_name('logits')]]))
        return inputs



# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
from dlk.utils.logger import Logger
import numpy as np
import pandas as pd
import os
import json

logger = Logger.get_logger()

@subprocessor_config_register('span_relation_relabel')
class SpanRelationRelabelConfig(BaseConfig):
    default_config = {
        "_name": "span_relation_relabel",
        "config": {
            "train": {
                "input_map": { 
                     "word_ids": "word_ids",
                     "offsets": "offsets",
                     "relations_info": "relations_info",
                     "processed_entities_info": "processed_entities_info",
                },
                "data_set": {
                     "train": ['train', 'valid', 'test'],
                },
                "output_map": {
                    "label_ids": "relation_label_ids",
                },
                "drop": "none", # 'longer'/'shorter'/'none', if entities is overlap, will remove by rule
                "vocab": "label_vocab#relation", # usually provided by the "token_gather" module
                "mask_fill": -100,
                "label_seperate": False, # seperate differences types of relations to different label matrix or use universal matrix(universal matrix may have conflict issue, like the 2 entities of the head of 2 different relations is same)
                "sym": True, # whther the from entity and end entity can swap in relations( if sym==True, we can just calc the upper trim and set down trim as -100 to ignore)
            },
            "extend_train": "train"
        }
    }
    f"""Config for SpanRelationRelabel
    Config Example:
        default_config
    """
    # {json.dumps(default_config, indent=4, ensure_ascii=False)}
    def __init__(self, stage, config: Dict):

        super(SpanRelationRelabelConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.offsets = self.config['input_map']['offsets']
        self.word_ids = self.config['input_map']['word_ids']
        self.mask_fill = self.config['mask_fill']
        self.sym = self.config['sym']
        self.label_seperate = self.config['label_seperate']
        self.relations_info = self.config['input_map']['relations_info']
        self.processed_entities_info = self.config['input_map']['processed_entities_info']
        self.vocab = self.config['vocab']
        self.output_labels = self.config['output_map']['label_ids']
        self.post_check(self.config, used=[
            "drop",
            "vocab",
            "mask_fill",
            "label_seperate",
            "sym",
            "input_map",
            "data_set",
            "output_map",
        ])


@subprocessor_register('span_relation_relabel')
class SpanRelationRelabel(ISubProcessor):
    """
    Relabel the json data to bio
    """

    def __init__(self, stage: str, config: SpanRelationRelabelConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        self.vocab = None
        if not self.data_set:
            logger.info(f"Skip 'span_relation_relabel' at stage {self.stage}")
            return

    def process(self, data: Dict)->Dict:
        """SpanRelationRelabel Entry

        Args:
            data: Dict

        Returns: 
            
            relabeled data

        """

        if not self.data_set:
            return data
        # NOTE: only load once, because the vocab should not be changed in same process
        if not self.vocab:
            self.vocab = Vocabulary.load(data[self.config.vocab])
            assert self.vocab.word2idx[self.vocab.unknown] == 0, f"For span_relation_relabel, 'unknown' must be index 0, and other labels as 1...num_label"
            assert not self.vocab.pad, f"For span_relation_relabel, 'pad' must be index 0, and other labels as 1...num_label"

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do span_relation_relabel on it.')
                continue
            data_set = data['data'][data_set_name]
            if os.environ.get('DISABLE_PANDAS_PARALLEL', 'false') != 'false':
                data_set[self.config.output_labels] = data_set.parallel_apply(self.relabel, axis=1)
            else:
                data_set[self.config.output_labels] = data_set.apply(self.relabel, axis=1)

        return data

    def relabel(self, one_ins: pd.Series):
        """make token label, if use the first piece label please use the 'span_cls_firstpiece_relabel'

        Args:
            one_ins: include sentence, entity_info, offsets

        Returns: 
            labels(labels for each subtoken)

        """
        relations_info = one_ins[self.config.relations_info]
        processed_entities_info = one_ins[self.config.processed_entities_info]
        entities_id_info_map = {}
        for processed_entity_info in processed_entities_info:
            entities_id_info_map[processed_entity_info['entity_id']] = processed_entity_info
        sub_word_ids = one_ins[self.config.word_ids]
        offsets = one_ins[self.config.offsets]
        offset_length = len(offsets)

        unknown_id = self.vocab.get_index(self.vocab.unknown)
        unknown_matrix = np.full((offset_length, offset_length), unknown_id)
        if self.config.sym:
            mask_matrix = np.full((offset_length, offset_length), self.config.mask_fill)
            mask_matrix = np.tril(mask_matrix, k=-1)
            unknown_matrix = np.triu(unknown_matrix, k=0)
            label_matrix = unknown_matrix + mask_matrix
        else:
            label_matrix = unknown_matrix
        if sub_word_ids[0] is None:
            label_matrix[0, :] = self.config.mask_fill
        if sub_word_ids[-1] is None:
            label_matrix[:, -1] = self.config.mask_fill
        if self.config.label_seperate:
            label_matrices = self.get_seperate_label_matrix(label_matrix, relations_info, entities_id_info_map)
        else:
            label_matrices = self.get_universal_label_matrix(label_matrix, relations_info, entities_id_info_map)

        return label_matrices

    def _get_entities_index(self, relation_info, entities_id_info_map):
        """get the from entity head, to entity head, from entity tail, to entity tail
        """
        from_index = relation_info['from']
        to_index = relation_info['to']
        from_entity = entities_id_info_map[from_index]
        from_start_index = from_entity['sub_token_start']
        from_end_index = from_entity['sub_token_end']
        to_entity = entities_id_info_map[to_index]
        to_start_index = to_entity['sub_token_start']
        to_end_index = to_entity['sub_token_end']

        if self.config.sym:
            if from_start_index > to_start_index:
                from_start_index, to_start_index = to_start_index, from_start_index
            if from_end_index > to_end_index:
                from_end_index, to_end_index = to_end_index, from_end_index
        return from_start_index, to_start_index, from_end_index, to_end_index 

    def get_seperate_label_matrix(self, label_matrix, relations_info: List, entities_id_info_map: Dict[int, Dict]):
        """seperate different type of relations to different matrix(as one dimension)

        Args:
            label_matrices: matrix
            relations_info: like
                [
                    {
                        "labels": [
                            "belong_to"
                        ],
                        "from": 1, # index of entity
                        "to": 0,
                    }
                ]
            entities_id_info_map: generate by span_cls_relabel

        Returns: 
            label_id matrix

        """
        label_types = len(self.vocab) - 1
        label_matrix = np.expand_dims(label_matrix, 0).repeat(2*label_types, 0)

        for relation_info in relations_info:
            label_id = self.vocab.word2idx[relation_info['labels'][0]]
            from_start_index, to_start_index, from_end_index, to_end_index = self._get_entities_index(relation_info, entities_id_info_map)
            # NOTE: label_id ==0 is unknown
            label_matrix[(label_id-1)*2][from_start_index][to_start_index] = 1
            label_matrix[(label_id-1)*2+1][from_end_index][to_end_index] = 1
        return label_matrix

    def get_universal_label_matrix(self, label_matrix, relations_info: List, entities_id_info_map: Dict[int, Dict]):
        """different type of relations to one matrix

        Args:
            label_matrices: matrix
            relations_info: like
                [
                    {
                        "labels": [
                            "belong_to"
                        ],
                        "from": 1, # index of entity
                        "to": 0,
                    }
                ]
            entities_id_info_map: generate by span_cls_relabel

        Returns:
            label_id matrix
        """
        label_matrix = np.expand_dims(label_matrix, 0).repeat(2, 0)
        for relation_info in relations_info:
            label_id = self.vocab.word2idx[relation_info['labels'][0]]
            from_start_index, to_start_index, from_end_index, to_end_index = self._get_entities_index(relation_info, entities_id_info_map)
            label_matrix[0][from_start_index][to_start_index] = label_id
            label_matrix[1][from_end_index][to_end_index] = label_id
        return label_matrix



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
from . import callback_register, callback_config_register
from typing import Dict, List
import os
from pytorch_lightning.callbacks import ModelCheckpoint


@callback_config_register('checkpoint')
class CheckpointCallbackConfig(object):
    """Config for CheckpointCallback

    Config Example:
        >>> {
        >>>     // default checkpoint configure
        >>>     "_name": "checkpoint",
        >>>     "config": {
        >>>         "monitor": "*@*",    // monitor which metrics or log value
        >>>         "save_top_k": 3,   //save top k
        >>>         "mode": "*@*", //"max" or "min" select topk min or max checkpoint, min for loss, max for acc
        >>>         "save_last": true,  //  always save last checkpoint
        >>>         "auto_insert_metric_name": true, //the save file name with or not metric name
        >>>         "every_n_train_steps": null, // Number of training steps between checkpoints.
        >>>         "every_n_epochs": 1, //Number of epochs between checkpoints.
        >>>         "save_on_train_epoch_end": false,// Whether to run checkpointing at the end of the training epoch. If this is False, then the check runs at the end of the validation.
        >>>         "save_weights_only": false, //whether save other status like optimizer, etc.
        >>>     }
        >>> }
    """
    def __init__(self, config: Dict):
        super(CheckpointCallbackConfig, self).__init__()
        config = config['config']
        self.monitor = config['monitor']
        self.save_last = config['save_last']
        self.save_top_k = config['save_top_k']
        self.mode = config['mode']
        self.auto_insert_metric_name = config['auto_insert_metric_name']
        self.every_n_train_steps = config['every_n_train_steps']
        self.every_n_epochs = config['every_n_epochs']
        self.save_on_train_epoch_end = config['save_on_train_epoch_end']
        self.save_weights_only = config['save_weights_only']

@callback_register('checkpoint')
class CheckpointCallback(object):
    """Save checkpoint decided by config
    """

    def __init__(self, config: CheckpointCallbackConfig):
        super().__init__()
        self.config = config

    def __call__(self, rt_config: Dict)->ModelCheckpoint:
        """get the ModelCheckpoint object

        Args:
            rt_config: runtime config, include save_dir, and the checkpoint path name

        Returns: 
            ModelCheckpoint object

        """
        dirpath = os.path.join(rt_config.get('save_dir', ''), rt_config.get("name", ''))
        return ModelCheckpoint(dirpath=dirpath, **self.config.__dict__)



# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from logging import PercentStyle
import hjson
from typing import Dict, Union, Callable, List
import torch
from dlk.core.models import model_register, model_config_register
from dlk.core.optimizers import optimizer_register, optimizer_config_register
from dlk.core.schedulers import scheduler_register, scheduler_config_register, BaseSchedulerConfig
from dlk.core.losses import loss_register, loss_config_register
from dlk.data.postprocessors import postprocessor_register, postprocessor_config_register
from dlk.utils.config import BaseConfig, ConfigTool
from . import imodel_config_register, imodel_register, GatherOutputMixin
from dlk.utils.logger import Logger
from functools import lru_cache
import copy
logger = Logger.get_logger()

import pytorch_lightning as pl

@imodel_config_register('basic')
class BasicIModelConfig(BaseConfig):
    default_config = {
        "_name": "basic",
        "model": "*@*",
        "loss": "*@*",
        "optimizer": {
            "_base": "adamw@bias_nodecay",
        },
        "schedule": {
            "_base": "linear_warmup"
        },
        "postprocessor": {
            "_base": "identity"
        },
        "config": {
        },
    }
    """ basic imodel config will provide all the config for model/optimizer/loss/scheduler/postprocess
    """
    def __init__(self, config: Dict):
        super(BasicIModelConfig, self).__init__(config)
        self.model, self.model_config = self.get_model(config.pop("model"))
        self.loss, self.loss_config = self.get_loss(config.pop("loss"))

        self.optimizer, self.optimizer_config = self.get_optimizer(config.pop("optimizer", 'adamw'))

        self.scheduler, self.scheduler_config = self.get_scheduler(config.pop("scheduler", "basic"))
        self.scheduler_config: BaseSchedulerConfig
        self.postprocess, self.postprocess_config = self.get_postprocessor(config.pop("postprocessor", 'identity'))

        self.config = config.pop('config', {})

    def get_postprocessor(self, config: Dict):
        """Use config to init the postprocessor

        Args:
            config: postprocess config

        Returns: 
            PostProcess, PostProcessConfig

        """
        return  ConfigTool.get_leaf_module(postprocessor_register, postprocessor_config_register, 'postprocessor', config)

    def get_model(self, config: Dict):
        """Use config to init the model

        Args:
            config: model config

        Returns: 
            Model, ModelConfig

        """
        return ConfigTool.get_leaf_module(model_register, model_config_register, "model", config)

    def get_loss(self, config: Dict):
        """Use config to init the loss

        Args:
            config: loss config

        Returns: 
            Loss, LossConfig

        """
        return ConfigTool.get_leaf_module(loss_register, loss_config_register, "loss", config)

    def get_optimizer(self, config: Dict):
        """Use config to init the optimizer

        Args:
            config: optimizer config

        Returns: 
            Optimizer, OptimizerConfig

        """
        return ConfigTool.get_leaf_module(optimizer_register, optimizer_config_register, "optimizer", config)

    def get_scheduler(self, config: Dict):
        """Use config to init the scheduler

        Args:
            config: scheduler config

        Returns: 
            Scheduler, SchedulerConfig

        """
        return ConfigTool.get_leaf_module(scheduler_register, scheduler_config_register, "scheduler", config)


@imodel_register("basic")
class BasicIModel(pl.LightningModule, GatherOutputMixin):
    """
    """
    def __init__(self, config: BasicIModelConfig, checkpoint=False):
        """ init all modules except scheduler which requires the information from datamodule(training steps and every epoch steps)
        """
        super().__init__()
        self.config = config  # scheduler will init in configure_optimizers, because it will use the datamodule info

        self.model = config.model(config.model_config, checkpoint)


        self.calc_loss = config.loss(config.loss_config)
        self.get_optimizer = config.optimizer(model=self.model, config=config.optimizer_config)

        self._origin_valid_data = None
        self._origin_test_data = None
        self.postprocessor = config.postprocess(config.postprocess_config)
        self.gather_data: Dict = copy.deepcopy(config.postprocess_config.input_map)
        self.gather_data.update(config.postprocess_config.predict_extend_return)

    def get_progress_bar_dict(self):
        """rewrite the prograss_bar_dict, remove the 'v_num' which we don't need

        Returns: 
            progress_bar dict

        """
        tqdm_dict = super().get_progress_bar_dict()
        tqdm_dict.pop("v_num", None)
        return tqdm_dict

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do forward on a mini batch

        Args:
            batch: a mini batch inputs

        Returns: 
            the outputs

        """
        return self.model(inputs)

    def training_step(self, batch: Dict[str, torch.Tensor], batch_idx: int):
        """do training_step on a mini batch

        Args:
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch

        Returns: 
            the outputs

        """
        result = self.model.training_step(batch)
        loss, loss_log = self.calc_loss(result, batch, rt_config={
            "current_step": self.global_step,
            "current_epoch": self.current_epoch,
            "total_steps": self.num_training_steps,
            "total_epochs": self.num_training_epochs
        })
        log_info = {}
        for key in loss_log:
            log_info[f"train_{key}"] = loss_log[key].unsqueeze(0)
        self.log_dict(log_info, prog_bar=True)
        return loss

    def validation_step(self, batch: Dict[str, torch.Tensor], batch_idx: int)->Dict[str, torch.Tensor]:
        """do validation on a mini batch

        The outputs only gather the keys in self.gather_data.keys for postprocess
        Args:
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch

        Returns: 
            the outputs

        """
        result = self.model.validation_step(batch)
        loss, loss_log = self.calc_loss(result, batch, rt_config={  # align with training step
            "current_step": self.global_step,
            "current_epoch": self.current_epoch,
            "total_steps": self.num_training_steps,
            "total_epochs": self.num_training_epochs
        })
        log_info = {}
        for key in loss_log:
            log_info[f"{key}"] = loss_log[key].unsqueeze(0)
        gather_column = list(self.gather_data.keys())
        return_result = log_info # this loss will be used in postprocess
        for column in gather_column:
            column = self.gather_data[column]
            if column in result:
                return_result[column] = result[column]
        return_result['_index'] = batch['_index']
        return return_result

    def validation_epoch_end(self, outputs: List[Dict])->List[Dict]:
        """Gather the outputs of all node and do postprocess on it.

        The outputs only gather the keys in self.gather_data.keys for postprocess
        Args:
            outputs: current node returnd output list

        Returns: 
            all node outputs

        """
        # BUG: the gather method is not implement well, so if you use ddp, the postprocessor will seperately run in deferent processores
        # outputs = self.gather_outputs(outputs)

        self.log_dict(
            self.postprocessor(stage='valid', list_batch_outputs=outputs, origin_data=self._origin_valid_data,
                rt_config={
                    "current_step": self.global_step,
                    "current_epoch": self.current_epoch,
                    "total_steps": self.num_training_steps,
                    "total_epochs": self.num_training_epochs
                }),
            prog_bar=True, rank_zero_only=True)
        return outputs

    def test_step(self, batch: Dict[str, torch.Tensor], batch_idx: int)->Dict:
        """do test on a mini batch

        The outputs only gather the keys in self.gather_data.keys for postprocess
        Args:
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch

        Returns: 
            the outputs

        """
        result = self.model.test_step(batch)
        loss, loss_log = self.calc_loss(result, batch, rt_config={  # align with training step
            "current_step": self.global_step,
            "current_epoch": self.current_epoch,
            "total_steps": self.num_training_steps,
            "total_epochs": self.num_training_epochs
        })
        log_info = {}
        for key in loss_log:
            log_info[f"{key}"] = loss_log[key].unsqueeze(0)
        gather_column = list(self.gather_data.keys())
        return_result = log_info # this loss will use in postprocess
        for column in gather_column:
            if column in result:
                return_result[column] = result[column]
        return_result['_index'] = batch['_index']
        return return_result

    def test_epoch_end(self, outputs: List[Dict])->List[Dict]:
        """Gather the outputs of all node and do postprocess on it.

        Args:
            outputs: current node returnd output list

        Returns: 
            all node outputs

        """
        # BUG: the gather method is not implement well, so if you use ddp, the postprocessor will seperately run in deferent processores
        # outputs = self.gather_outputs(outputs)

        self.log_dict(
            self.postprocessor(stage='test', list_batch_outputs=outputs, origin_data=self._origin_test_data,
                rt_config={
                    "current_step": self.global_step,
                    "current_epoch": self.current_epoch,
                    "total_steps": self.num_training_steps,
                    "total_epochs": self.num_training_epochs
                }),
            prog_bar=True, rank_zero_only=True)
        return outputs

    def predict_step(self, batch: Dict, batch_idx: int)->Dict:
        """do predict on a mini batch

        Args:
            batch: a mini batch inputs
            batch_idx: the index(dataloader) of the mini batch

        Returns: 
            the outputs

        """
        result = self.model.predict_step(batch)
        gather_column = list(self.gather_data.keys())
        return_result = {}
        for column in gather_column:
            column = self.gather_data[column]
            if column in result:
                return_result[column] = result[column]
        return_result['_index'] = batch['_index']
        return return_result

    @property
    @lru_cache(maxsize=5) # the size should always == 1
    def num_training_epochs(self) -> int:
        """Total training epochs inferred from datamodule and devices.
        """
        return self.trainer.max_epochs

    @property
    @lru_cache(maxsize=2) # the size should always == 1
    def epoch_training_steps(self) -> int:
        """every epoch training steps inferred from datamodule and devices.
        """
        if self.trainer.datamodule.train_dataloader() is None:
            batches = 0
        else:
            batches = len(self.trainer.datamodule.train_dataloader())
        return batches

    @property
    @lru_cache(maxsize=2) # the size should always == 1
    def num_training_steps(self) -> int:
        """Total training steps inferred from datamodule and devices.
        """
        if self.trainer.max_steps != -1:
            return self.trainer.max_steps
        # FIXED: https://github.com/PyTorchLightning/pytorch-lightning/pull/11599
        # NEED TEST
        return self.trainer.estimated_stepping_batches
        # legacy version pl<=1.5.8
        # FIXIT: https://github.com/PyTorchLightning/pytorch-lightning/issues/5449 should check update
        #        
        # limit_batches = self.trainer.limit_train_batches
        # if self.trainer.datamodule.train_dataloader() is None:
        #     batches = 0
        # else:
        #     batches = len(self.trainer.datamodule.train_dataloader())
        # batches = min(batches, limit_batches) if isinstance(limit_batches, int) else int(limit_batches * batches)

        # num_devices = max(1, self.trainer.devices)
        # if self.trainer.num_devices:
        #     num_devices = max(num_devices, self.trainer.num_devices)

        # effective_accum = self.trainer.accumulate_grad_batches * num_devices
        # return (batches // effective_accum + (1 if batches%effective_accum else 0)) * self.trainer.max_epochs

    def configure_optimizers(self):
        """Configure the optimizer and scheduler
        """

        self.calc_loss.update_config(rt_config={
            "total_steps": self.num_training_steps,
            "total_epochs": self.num_training_epochs
        })

        optimizer = self.get_optimizer()
        self.config.scheduler_config.num_training_steps = self.num_training_steps
        self.config.scheduler_config.epoch_training_steps = self.epoch_training_steps
        self.config.scheduler_config.num_training_epochs = self.num_training_epochs
        self.config.scheduler_config.last_epoch = self.num_training_epochs
        scheduler = self.config.scheduler(optimizer, self.config.scheduler_config)()

        return {
            "optimizer": optimizer,
            "lr_scheduler": {
                "scheduler": scheduler,
                "interval": "step",
                "frequency": 1
            }
        }



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dlk.utils.config import ConfigTool, BaseConfig
from dlk.utils.logger import Logger
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from dlk.utils.io import open
import pickle as pkl
import os

logger = Logger.get_logger()


@subprocessor_config_register('load')
class LoadConfig(BaseConfig):
    default_config = {
        "_name": "load",
        "config": {
            "base_dir": "",
            "predict": {
                "meta": "meta.pkl",
            },
            "online": "predict",
            "extend_train": "predict"
        }
    }
    """Config for Load

    Config Example:
        default_config
    """
    def __init__(self, stage, config):
        super(LoadConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.base_dir: str = config.get('config').get("base_dir", "")


@subprocessor_register('load')
class Load(ISubProcessor):
    """ Loader the $meta, etc. to data
    """
    def __init__(self, stage: str, config: LoadConfig):
        super().__init__()
        self.stage = stage
        self.config = config.config
        self.load_data = {}
        if not self.config:
            logger.info(f"Skip 'load' at stage {self.stage}")
            return
        self.base_dir = config.base_dir

        for key, path in self.config.items():
            self.load_data[key] = self.load(path)

    def load(self, path: str):
        """load data from path

        Args:
            path: the path to data

        Returns: 
            loaded data

        """

        logger.info(f"Loading file from {os.path.join(self.base_dir, path)}")
        with open(os.path.join(self.base_dir, path), 'rb') as f:
            return pkl.load(f)

    def process(self, data: Dict) -> Dict:
        """Load entry

        Args:
            data: 
            >>> {
            >>>     "data": {"train": ...},
            >>>     "tokenizer": ..
            >>> }

        Returns: 
            data + loaded_data

        """
        for _, meta in self.load_data.items():
            for key, value in meta.items():
                data[key] = value
        return data



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numpy import result_type
from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
from dlk.utils.logger import Logger
import os
import pandas as pd

logger = Logger.get_logger()

@subprocessor_config_register('word_mask')
class WordMaskConfig(BaseConfig):
    default_config = {
        "_name": "word_mask",
        "config": {
            "train":{
                "input_map": {  # without necessery, don't change this
                    "word_ids": "word_ids",
                    "offsets": "offsets",
                    "entities_info": "entities_info",
                },
                "data_set": {                   # for different stage, this processor will process different part of data
                    "train": ['train', 'valid', 'test'],
                    "predict": ['predict'],
                    "online": ['online']
                },
                "output_map": {
                    "labels": "labels",
                },
            },
            "predict": "train",
            "extend_train": "train",
            "online": "train",
        }
    }
    """Config for WordMask

    Config Example:
        default_config
    """
    def __init__(self, stage, config: Dict):

        super(WordMaskConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.word_ids = self.config['input_map']['word_ids']
        self.offsets = self.config['input_map']['offsets']
        self.entities_info = self.config['input_map']['entities_info']
        self.clean_droped_entity = self.config['clean_droped_entity']
        self.drop = self.config['drop']
        self.start_label = self.config['start_label']
        self.end_label = self.config['end_label']
        self.output_labels = self.config['output_map']['labels']
        self.entity_priority = {entity: priority for priority, entity in enumerate(self.config['entity_priority'])}
        self.priority_trigger = self.config['priority_trigger']
        self.post_check(self.config, used=[
            "input_map",
            "data_set",
            "drop",
            "output_map",
        ])


@subprocessor_register('word_mask')
class WordMask(ISubProcessor):
    """
    Relabel the json data to bio
    """

    def __init__(self, stage: str, config: WordMaskConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'word_mask' at stage {self.stage}")
            return

    def process(self, data: Dict)->Dict:
        """WordMask Entry

        Args:
            data: Dict

        Returns: 
            
            relabeled data

        """

        if not self.data_set:
            return data

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do word_mask on it.')
                continue
            data_set = data['data'][data_set_name]
            if os.environ.get('DISABLE_PANDAS_PARALLEL', 'false') != 'false':
                data_set[[self.config.output_labels,
                    self.config.entities_info
                ]] = data_set.parallel_apply(self.relabel, axis=1, result_type='expand')
            else:
                data_set[[self.config.output_labels,
                    self.config.entities_info
                ]] = data_set.apply(self.relabel, axis=1, result_type='expand')

        return data



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ahocorasick
from typing import Iterable, List, Dict

class QuickSearch(object):
    """Ahocorasick enhanced Trie"""
    def __init__(self, words: Iterable=[]):
        """init the QuickSearch

        Args:
            words: tokens in use words to update the trie

        Returns: 
            None

        """
        super(QuickSearch, self).__init__()

        if not isinstance(words, Iterable) :
            raise PermissionError("The words must be a Iterable object.")
        self.next_index = 0

        self.ac = ahocorasick.Automaton()

        for sub_str in words:
            sub_str = f" {sub_str.strip()} "
            self.ac.add_word(sub_str, (self.next_index, sub_str))
            self.next_index += 1
        self.ac.make_automaton()

    def search(self, search_str:str)->List[Dict]:
        """find whether some sub_str in trie

        Args:
            search_str: find the search_str

        Returns: 
            >>> list of result:
            >>> the result organized as {
            >>>     "start": start_position,
            >>>     "end": end_position,
            >>>     "str": search_str[start_position: end_position]
            >>> }

        """
        result = []
        for end_index, (_, original_value) in self.ac.iter(search_str):
            start_index = end_index - len(original_value) + 1
            end_index = end_index + 1
            result.append({"start": start_index, "end": end_index, "str": original_value})
        return result

    def has(self, key: str)->bool:
        """check key is in trie

        Args:
            key: a token(str)

        Returns: 
            bool(has or not)

        """
        try:
            self.ac.get(key)
        except:
            return False
        return True

    def add_word(self, word: str):
        """add a single word to trie

        Args:
            word: single token

        Returns: 
            None

        """
        self.ac.add_word(word, (self.next_index, word))
        self.next_index += 1
        self.ac.make_automaton()

    def add_words(self, words: Iterable):
        """add words from iterator to the trie

        Args:
            words: Iterable[tokens]

        Returns: 
            None

        """
        for sub_str in words:
            self.ac.add_word(sub_str, (self.next_index, sub_str))
            self.next_index += 1
        self.ac.make_automaton()



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from typing import Dict, List, Set, Callable
from dlk.core.base_module import BaseModule, BaseModuleConfig
from . import decoder_register, decoder_config_register
from dlk.core.modules import module_config_register, module_register
import copy

@decoder_config_register("linear_crf")
class LinearCRFConfig(BaseModuleConfig):
    """Config for LinearCRF 

    Config Example:
        >>> {
        >>>     "module@linear": {
        >>>         "_base": "linear",
        >>>     },
        >>>     "module@crf": {
        >>>         "_base": "crf",
        >>>     },
        >>>     "config": {
        >>>         "input_size": "*@*",  // the linear input_size
        >>>         "output_size": "*@*", // the linear output_size
        >>>         "reduction": "mean", // crf reduction method
        >>>         "output_map": {}, //provide_key: output_key
        >>>         "input_map": {} // required_key: provide_key
        >>>     },
        >>>     "_link":{
        >>>         "config.input_size": ["module@linear.config.input_size"],
        >>>         "config.output_size": ["module@linear.config.output_size", "module@crf.config.output_size"],
        >>>         "config.reduction": ["module@crf.config.reduction"],
        >>>     }
        >>>     "_name": "linear_crf",
        >>> }
    """
    def __init__(self, config: Dict):
        super(LinearCRFConfig, self).__init__(config)
        self.linear_config = config["module@linear"]
        self.crf_config = config["module@crf"]
        self.post_check(config['config'], used=[
                            'input_size',
                            'output_size',
                            'reduction',
                            "return_logits",
                        ])


@decoder_register("linear_crf")
class LinearCRF(BaseModule):
    """use torch.nn.Linear get the emission probability and fit to CRF"""
    def __init__(self, config: LinearCRFConfig):
        super(LinearCRF, self).__init__(config)
        self._provide_keys = {'logits', "predict_seq_label"}
        self._required_keys = {'embedding', 'label_ids', 'attention_mask'}

        self.config = config
        self.linear = module_register.get('linear')(module_config_register.get('linear')(config.linear_config))
        self.crf = module_register.get('crf')(module_config_register.get('crf')(config.crf_config))

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns: 
            None

        """
        self.linear.init_weight(method)
        self.crf.init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do predict, only get the predict labels

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        return self.predict_step(inputs)

    def predict_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do predict, only get the predict labels

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        logits = self.linear(inputs[self.get_input_name('embedding')])
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([logits]))
        inputs[self.get_output_name("predict_seq_label")] = self.crf(logits, inputs[self.get_input_name('attention_mask')])
        return inputs

    def training_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do training step, get the crf loss

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        logits = self.linear(inputs[self.get_input_name('embedding')])
        loss = self.crf.training_step(logits, inputs[self.get_input_name('label_ids')], inputs[self.get_input_name('attention_mask')])
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([logits]))
        inputs[self.get_output_name('loss')] = loss
        return inputs

    def validation_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do validation step, get the crf loss and the predict labels

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        logits = self.linear(inputs[self.get_input_name('embedding')])
        loss = self.crf.training_step(logits, inputs[self.get_input_name('label_ids')], inputs[self.get_input_name('attention_mask')])
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([logits]))
        inputs[self.get_output_name('loss')] = loss
        inputs[self.get_output_name("predict_seq_label")] = self.crf(logits, inputs[self.get_input_name('attention_mask')])
        return inputs



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http:# www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
import numpy as np
import torch
from typing import Dict, List, Optional
from . import module_register, module_config_register, Module
from dlk.utils.config import BaseConfig
from dlk.utils.logger import Logger

logger = Logger.get_logger()

class RoFormerSinusoidalPositionalEmbedding(nn.Embedding):
    """This module produces sinusoidal positional embeddings of any length."""

    def __init__(self, num_positions: int, embedding_dim: int, padding_idx: Optional[int] = None) -> None:
        super().__init__(num_positions, embedding_dim)
        self.weight = self._init_weight(self.weight)

    @staticmethod
    def _init_weight(out: nn.Parameter) -> nn.Parameter:
        """
        Identical to the XLM create_sinusoidal_embeddings except features are not interleaved. The cos features are in
        the 2nd half of the vector. [dim // 2:]
        """
        n_pos, dim = out.shape
        position_enc = np.array(
            [[pos / np.power(10000, 2 * (j // 2) / dim) for j in range(dim)] for pos in range(n_pos)]
        )
        out.requires_grad = False  # set early to avoid an error in pytorch-1.8+
        sentinel = dim // 2 if dim % 2 == 0 else (dim // 2) + 1
        out[:, 0:sentinel] = torch.FloatTensor(np.sin(position_enc[:, 0::2]))
        out[:, sentinel:] = torch.FloatTensor(np.cos(position_enc[:, 1::2]))
        out.detach_()
        return out

    @torch.no_grad()
    def forward(self, input_ids_shape: torch.Size, past_key_values_length: int = 0) -> torch.Tensor:
        """`input_ids_shape` is expected to be [bsz x seqlen]."""
        bsz, seq_len = input_ids_shape[:2]
        positions = torch.arange(
            past_key_values_length, past_key_values_length + seq_len, dtype=torch.long, device=self.weight.device
        )
        return super().forward(positions)

    @staticmethod
    def apply_rotary_position_embeddings(sinusoidal_pos, query_layer, key_layer, value_layer=None):
        # https://kexue.fm/archives/8265
        # sin [batch_size, num_heads, sequence_length, embed_size_per_head//2]
        # cos [batch_size, num_heads, sequence_length, embed_size_per_head//2]
        sin, cos = sinusoidal_pos.chunk(2, dim=-1)
        # sin [θ0,θ1,θ2......θd/2-1] -> sin_pos [θ0,θ0,θ1,θ1,θ2,θ2......θd/2-1,θd/2-1]
        sin_pos = torch.stack([sin, sin], dim=-1).reshape_as(sinusoidal_pos)
        # cos [θ0,θ1,θ2......θd/2-1] -> cos_pos [θ0,θ0,θ1,θ1,θ2,θ2......θd/2-1,θd/2-1]
        cos_pos = torch.stack([cos, cos], dim=-1).reshape_as(sinusoidal_pos)
        # rotate_half_query_layer [-q1,q0,-q3,q2......,-qd-1,qd-2]
        rotate_half_query_layer = torch.stack([-query_layer[..., 1::2], query_layer[..., ::2]], dim=-1).reshape_as(
            query_layer
        )
        query_layer = query_layer * cos_pos + rotate_half_query_layer * sin_pos
        # rotate_half_key_layer [-k1,k0,-k3,k2......,-kd-1,kd-2]
        rotate_half_key_layer = torch.stack([-key_layer[..., 1::2], key_layer[..., ::2]], dim=-1).reshape_as(key_layer)
        key_layer = key_layer * cos_pos + rotate_half_key_layer * sin_pos
        if value_layer is not None:
            # rotate_half_value_layer [-v1,v0,-v3,v2......,-vd-1,vd-2]
            rotate_half_value_layer = torch.stack([-value_layer[..., 1::2], value_layer[..., ::2]], dim=-1).reshape_as(
                value_layer
            )
            value_layer = value_layer * cos_pos + rotate_half_value_layer * sin_pos
            return query_layer, key_layer, value_layer
        return query_layer, key_layer

@module_config_register("biaffine")
class BiAffineConfig(BaseConfig):
    default_config = {
        "_name": "biaffine",
        "config": {
            "input_size": "*@*",
            "hidden_size": 0, # default == input_size
            "output_size": "*@*",
            "dropout": 0.0, # generally no need dropout
            "multi_matrix": 1, # like relation need head pair and tail pair calc togather, so the multi_matrix should set to >1
            "relation_position": False, # whether add relation_position before align
            "max_seq_len": 1024, # for relation_position
            "bias": True, # use bias or not in biaffine
        },
    }
    """Config for BiAffine
    Config Example:
    """
    def __init__(self, config: Dict):
        super(BiAffineConfig, self).__init__(config)
        config = config['config']
        self.input_size = config['input_size']
        self.multi_matrix = config['multi_matrix']
        self.relation_position = config['relation_position']
        self.target_size = config['output_size']
        self.max_seq_len = config['max_seq_len']
        self.output_size = config['output_size'] * self.multi_matrix
        self.hidden_size = config['hidden_size']
        if not self.hidden_size:
            self.hidden_size = self.input_size
        self.dropout = config['dropout']
        self.dropout = float(config['dropout'])
        self.bias = config['bias']
        self.post_check(config, used=[
            "input_size",
            "hidden_size",
            "output_size",
            "dropout",
            "multi_matrix",
            "relation_position",
            "max_seq_len",
            "bias",
        ])


@module_register("biaffine")
class BiAffine(Module):
    """wrap for nn.BiAffine"""
    def __init__(self, config: BiAffineConfig):
        super(BiAffine, self).__init__()
        self.linear_a = nn.Linear(config.input_size, config.hidden_size)
        self.linear_b = nn.Linear(config.input_size, config.hidden_size)
        self.config = config
        if config.relation_position:
            self.embed_positions = RoFormerSinusoidalPositionalEmbedding(
                config.max_seq_len, config.hidden_size
            )
        self.dropout = nn.Dropout(p=config.dropout)
        self.active = nn.LeakyReLU() # TODO: why GELU get loss nan?
        if config.bias:
            self.biaffine = nn.Parameter(torch.randn(config.hidden_size+1, config.output_size, config.hidden_size+1))
        else:
            self.biaffine = nn.Parameter(torch.randn(config.hidden_size, config.output_size, config.hidden_size))

        self.dropout = nn.Dropout(p=float(config.dropout))
        self.config = config

    def init_weight(self, method):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns: 
            None

        """
        torch.nn.init.xavier_uniform_(self.biaffine)
        self.linear_a.apply(method)
        self.linear_b.apply(method)

    def forward(self, embedding: torch.Tensor)->torch.Tensor:
        """do forward on a mini batch

        Args:
            embedding: a mini batch embedding, shape==(batch_size, input_a_len, input_size)

        Returns: 
            input_a x biaffine x input_b, shape==(batch_size, input_a_len, input_b_len, output_size)

        """

        input_a = self.dropout(self.active(embedding))
        input_b = self.dropout(self.active(embedding))
        input_a = self.linear_a(input_a)
        input_b = self.linear_b(input_b)

        if self.config.relation_position:
            sinusoidal_pos = self.embed_positions(input_a.shape[:-1])[None, :, :]
            input_a, input_b = self.embed_positions.apply_rotary_position_embeddings(sinusoidal_pos, input_a, input_b)

        if self.config.bias:
            output = torch.einsum('bmi,ioj,bnj->bmno', 
                    torch.cat((input_a, torch.ones_like(input_a[..., :1])), dim=-1), 
                    self.biaffine, 
                    torch.cat((input_b, torch.ones_like(input_b[..., :1])), dim=-1)
                    )
        else:
            output = torch.einsum('bmi,ioj,bnj->bmno', 
                    input_a,
                    self.biaffine, 
                    input_b,
                    )
        if self.config.multi_matrix>1:
            bs, seq_len, _, output_size = output.shape
            output = output.reshape(bs, seq_len, seq_len, self.config.multi_matrix, self.config.target_size)
            output = output.permute(0, 3, 1, 2, 4) # bs, group, seq_len, seq_len, target_size)
        return output



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""postprocessors"""

import importlib
import os
from typing import Callable, Dict, Type, List, Union
from dlk.utils.register import Register
from dlk.utils.config import BaseConfig
import torch
import pandas as pd
import abc

class IPostProcessorConfig(BaseConfig):
    """docstring for PostProcessorConfigBase"""
    def __init__(self, config):
        super(IPostProcessorConfig, self).__init__(config)
        self.config = config.get('config', {})

    @property
    def predict_extend_return(self):
        """save the extend data in predict

        Returns: 
            predict_extend_return
        """
        return self.config.get('predict_extend_return', {})

    @property
    def input_map(self):
        """required the output of model process content name map

        Returns: 
            input_map
        """
        return self.config.get("input_map", {})

    @property
    def origin_input_map(self):
        """required the origin data(before pass to datamodule) column name map

        Returns: 
            origin_input_map
        """
        return self.config.get("origin_input_map", {})


class IPostProcessor(object):
    """docstring for IPostProcessor"""

    def __init__(self, config):
        super(IPostProcessor, self).__init__()
        self.config = config

    def loss_name_map(self, stage)->str:
        """get the stage loss name

        Args:
            stage: valid, train or test

        Returns: 
            loss_name

        """
        map = {
            "valid": 'val',
            'train': 'train',
            "test": "test",
        }
        return map.get(stage, stage)

    def gather_predict_extend_data(self, input_data: Dict, i: int, predict_extend_return: Dict):
        """gather the data register in `predict_extend_return`
        Args:
            input_data:
                the model output
            i:
                the index is i
            predict_extend_return: 
                the name map which will be reserved
        Returns: 
            a dict of data in input_data which is register in predict_extend_return
        """
        result = {}
        for key, name in predict_extend_return.items():
            data = input_data[name][i]
            if torch.is_tensor(data):
                data = data.detach().tolist()
            result[key] = data
        return result

    def average_loss(self, list_batch_outputs: List[Dict])->Dict[str, float]:
        """average all the loss of the list_batches

        Args:
            list_batch_outputs: a list of outputs

        Returns: 
            average_loss

        """
        loss_names = []
        average_losses = {}
        batch_num = len(list_batch_outputs)
        if not batch_num:
            return average_losses
        for key in list_batch_outputs[0]:
            if key.endswith("_loss") or key == 'loss':
                loss_names.append(key)
                average_losses[key] = 0
        for batch_output in list_batch_outputs:
            for name in loss_names:
                average_losses[name] = average_losses[name] + batch_output.get(name, 0)
        average_losses = {key: value/batch_num for key, value in average_losses.items()}
        return average_losses

    @abc.abstractmethod
    def do_predict(self, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict)->List:
        """Process the model predict to human readable format

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns: 
            all predicts

        """
        raise NotImplementedError

    @abc.abstractmethod
    def do_calc_metrics(self, predicts: List, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict)->Dict:
        """calc the scores use the predicts or list_batch_outputs

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> 
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns: 
            the named scores

        """
        raise NotImplementedError

    @abc.abstractmethod
    def do_save(self, predicts: List, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict, save_condition: bool=False):
        """save the predict when save_condition==True

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            save_condition: True for save, False for depend on rt_config

        Returns: 
            None

        """
        raise NotImplementedError

    @property
    def without_ground_truth_stage(self)->set:
        """there is not groud truth in the returned stage

        Returns: 
            without_ground_truth_stage

        """
        return {'predict', 'online'}

    def process(self, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict, save_condition: bool=False)->Union[Dict, List]:
        """PostProcess entry

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            save_condition: if save_condition is True, will force save the predict on all stage except online

        Returns: 
            the log_info(metrics) or the stage is "online" return the predicts

        """
        log_info = {}
        if stage not in self.without_ground_truth_stage:
            average_loss = self.average_loss(list_batch_outputs=list_batch_outputs)
            for name in average_loss:
                log_info[f'{self.loss_name_map(stage)}_{name}'] = average_loss[name]
        predicts = self.do_predict(stage, list_batch_outputs, origin_data, rt_config)
        if stage not in self.without_ground_truth_stage:
            log_info.update(self.do_calc_metrics(predicts, stage, list_batch_outputs, origin_data, rt_config))

        if stage == 'online':
            return predicts
        if stage == 'predict':
            if save_condition:
                self.do_save(predicts, stage, list_batch_outputs, origin_data, rt_config, save_condition=True)
            return predicts
        else:
            if save_condition:
                self.do_save(predicts, stage, list_batch_outputs, origin_data, rt_config, save_condition=True)
            else:
                self.do_save(predicts, stage, list_batch_outputs, origin_data, rt_config, save_condition=False)
            return log_info

    def __call__(self, stage, list_batch_outputs, origin_data, rt_config, save_condition=False):
        """the same as self.process
        """
        return self.process(stage, list_batch_outputs, origin_data, rt_config, save_condition)


postprocessor_config_register = Register('PostProcessor config register')
postprocessor_register = Register("PostProcessor register")


def import_postprocessors(postprocessors_dir, namespace):
    for file in os.listdir(postprocessors_dir):
        path = os.path.join(postprocessors_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            # and not (file.endswith("subpostprocessors") and os.path.isdir(path))
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            postprocessor_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + postprocessor_name)


# automatically import any Python files in the models directory
postprocessors_dir = os.path.dirname(__file__)
import_postprocessors(postprocessors_dir, "dlk.data.postprocessors")



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Get the dlk package root path
"""
import os

ROOT = os.path.join('/')

def get_root():
    """get the dlk root

    Returns: 
        abspath of this package

    """
    cur_dir = os.path.dirname(os.path.abspath(__file__))

    def check_cur_is_root(cur_dir):
        """check cur_dir has 'dlk' or not
        """
        root_sign_list = ['dlk']
        for root_sign in root_sign_list:
            if os.path.exists(os.path.join(cur_dir, root_sign)):
                return True
        return False

    while cur_dir != ROOT:
        if check_cur_is_root(cur_dir):
            break
        cur_dir = os.path.dirname(cur_dir)

    if cur_dir == ROOT:
        raise ImportError(f'{os.path.dirname(os.path.abspath(__file__))} is not a right file to find dlk root')
    else:
        return cur_dir



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
import math
from . import scheduler_register, scheduler_config_register, BaseScheduler, BaseSchedulerConfig
from torch.optim.lr_scheduler import LambdaLR
import torch.optim as optim


@scheduler_config_register("cosine_warmup")
class CosineWarmupScheduleConfig(BaseSchedulerConfig):
    """Config for CosineWarmupSchedule

    Config Example:
        >>> {
        >>>     "config": {
        >>>         "num_warmup_steps": 0,
        >>>         "num_cycles": 0.5,
        >>>     },
        >>>     "_name": "cosine_warmup",
        >>> }
    """
    def __init__(self, config: Dict):
        super(CosineWarmupScheduleConfig, self).__init__(config)
        config = config['config']
        self.num_warmup_steps = config["num_warmup_steps"]
        self.num_cycles = config['num_cycles']
        self.post_check(config, used=[
            "num_warmup_steps",
            "num_cycles",
        ])


@scheduler_register("cosine_warmup")
class CosineWarmupSchedule(BaseScheduler):
    """CosineWarmupSchedule"""
    def __init__(self, optimizer: optim.Optimizer, config: CosineWarmupScheduleConfig):
        super(CosineWarmupSchedule, self).__init__()
        self.config = config
        self.optimizer = optimizer

    def get_scheduler(self)->LambdaLR:
        """return the initialized linear wramup then cos decay scheduler

        Returns: 
            Schedule

        """
        num_training_steps = self.config.num_training_steps
        num_warmup_steps = self.config.num_warmup_steps
        if num_warmup_steps >0 and num_warmup_steps < 1:
            num_warmup_steps = int(num_warmup_steps * num_training_steps)

        num_cycles = self.config.num_cycles


        def lr_lambda(current_step):
            if current_step < num_warmup_steps:
                return float(current_step) / float(max(1, num_warmup_steps))
            progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))
            return max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)))

        return LambdaLR(self.optimizer, lr_lambda, last_epoch=-1)



# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numpy import result_type
from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
from dlk.utils.logger import Logger
import os
import pandas as pd

logger = Logger.get_logger()

@subprocessor_config_register('seq_lab_relabel')
class SeqLabRelabelConfig(BaseConfig):
    default_config = {
        "_name": "seq_lab_relabel",
        "config": {
            "train":{
                "input_map": {  # without necessery, don't change this
                    "word_ids": "word_ids",
                    "offsets": "offsets",
                    "entities_info": "entities_info",
                },
                "data_set": {                   # for different stage, this processor will process different part of data
                    "train": ['train', 'valid', 'test'],
                    "predict": ['predict'],
                    "online": ['online']
                },
                "output_map": {
                    "labels": "labels",
                },
                "drop": "shorter", #longer'/'shorter'/'none', if entities is overlap, will remove by rule
                "start_label": "S",
                "end_label": "E",
                "clean_droped_entity": True, # after drop entity for training, whether drop the entity for calc metrics, default is true, this only works when the drop != 'none'
                "entity_priority": [],
                #entity_priority": ['Product'],
                "priority_trigger": 1, # if the overlap entity abs(length_a - length_b)<=priority_trigger, will trigger the entity_priority strategy
            },
            "extend_train": "train",
        }
    }
    """Config for SeqLabRelabel

    Config Example:
        default_config
    """
    def __init__(self, stage, config: Dict):

        super(SeqLabRelabelConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.word_ids = self.config['input_map']['word_ids']
        self.offsets = self.config['input_map']['offsets']
        self.entities_info = self.config['input_map']['entities_info']
        self.clean_droped_entity = self.config['clean_droped_entity']
        self.drop = self.config['drop']
        self.start_label = self.config['start_label']
        self.end_label = self.config['end_label']
        self.output_labels = self.config['output_map']['labels']
        self.entity_priority = {entity: priority for priority, entity in enumerate(self.config['entity_priority'])}
        self.priority_trigger = self.config['priority_trigger']
        self.post_check(self.config, used=[
            "input_map",
            "data_set",
            "drop",
            "output_map",
            "start_label",
            "end_label",
            "clean_droped_entity",
            "entity_priority",
            "priority_trigger",
        ])


@subprocessor_register('seq_lab_relabel')
class SeqLabRelabel(ISubProcessor):
    """
    Relabel the json data to bio
    """

    def __init__(self, stage: str, config: SeqLabRelabelConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'seq_lab_relabel' at stage {self.stage}")
            return

    def process(self, data: Dict)->Dict:
        """SeqLabRelabel Entry

        Args:
            data: Dict

        Returns: 
            
            relabeled data

        """

        if not self.data_set:
            return data

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do seq_lab_relabel on it.')
                continue
            data_set = data['data'][data_set_name]
            if os.environ.get('DISABLE_PANDAS_PARALLEL', 'false') != 'false':
                data_set[[self.config.output_labels,
                    self.config.entities_info
                ]] = data_set.parallel_apply(self.relabel, axis=1, result_type='expand')
            else:
                data_set[[self.config.output_labels,
                    self.config.entities_info
                ]] = data_set.apply(self.relabel, axis=1, result_type='expand')

        return data

    def find_position_in_offsets(self, position: int, offset_list: List, sub_word_ids: List, start: int, end: int, is_start: bool=False):
        """find the sub_word index which the offset_list[index][0]<=position<offset_list[index][1]

        Args:
            position: position
            offset_list: list of all tokens offsets
            sub_word_ids: word_ids from tokenizer
            start: start search index
            end: end search index
            is_start: is the position is the start of target token, if the is_start==True and cannot find return -1

        Returns: 
            the index of the offset which include position

        """
        while start<end:
            if sub_word_ids[start] is None:
                start += 1
            elif position>=offset_list[start][0] and position<offset_list[start][1]:
                return start
            elif position<offset_list[start][0]:
                if start == 1 and offset_list[0] == [0, 0]:
                    return 1
                if is_start:
                    return -1
                else:
                    return start - 1
            else:
                start += 1
        return -1

    def relabel(self, one_ins: pd.Series):
        """make token label, if use the first piece label please use the 'seq_lab_firstpiece_relabel'

        Args:
            one_ins: include sentence, entity_info, offsets

        Returns: 
            labels(labels for each subtoken)

        """
        pre_clean_entities_info = one_ins[self.config.entities_info]
        pre_clean_entities_info.sort(key=lambda x: x['start'])
        offsets = one_ins[self.config.offsets]
        sub_word_ids = one_ins[self.config.word_ids]
        if not sub_word_ids:
            logger.warning(f"entity_info: {pre_clean_entities_info}, offsets: {offsets} ")

        entities_info = []
        pre_end = -1
        pre_length = 0
        pre_label = ''
        for entity_info in pre_clean_entities_info:
            assert len(entity_info['labels']) == 1, f"currently we just support one label for one entity"
            if entity_info['start']<pre_end: # if overlap will remove one
                if self.config.drop == 'none':
                    pass
                elif abs(entity_info['end'] - entity_info['start'] - pre_length) <= self.config.priority_trigger:
                    pre_label_order = self.config.entity_priority.get(pre_label, 1e9)
                    label_order = self.config.entity_priority.get(entity_info['labels'][0], 1e9)
                    if label_order<pre_label_order:
                        entities_info.pop()
                    else:
                        continue
                elif self.config.drop == 'shorter':
                    if entity_info['end'] - entity_info['start'] > pre_length:
                        entities_info.pop()
                    else:
                        continue
                elif self.config.drop =='longer':
                    if entity_info['end'] - entity_info['start'] < pre_length:
                        entities_info.pop()
                    else:
                        continue
                else:
                    raise PermissionError(f"The drop method must in 'none'/'shorter'/'longer'")
                pre_label = entity_info['labels'][0]
            entities_info.append(entity_info)
            pre_end = entity_info['end']
            pre_length = entity_info['end'] - entity_info['start']

        cur_token_index = 0
        offset_length = len(offsets)
        sub_labels = []
        for entity_info in entities_info:
            start_token_index = self.find_position_in_offsets(entity_info['start'], offsets, sub_word_ids, cur_token_index, offset_length, is_start=True)
            if start_token_index == -1:
                logger.warning(f"cannot find the entity_info : {entity_info}, offsets: {offsets} ")
                continue
            for _ in range(start_token_index-cur_token_index):
                sub_labels.append('O')
            end_token_index = self.find_position_in_offsets(entity_info['end']-1, offsets, sub_word_ids, start_token_index, offset_length)
            assert end_token_index != -1, f"entity_info: {entity_info}, offsets: {offsets}"
            sub_labels.append("B-"+entity_info['labels'][0])
            for _ in range(end_token_index-start_token_index):
                sub_labels.append("I-"+entity_info['labels'][0])
            cur_token_index = end_token_index + 1
        assert cur_token_index<=offset_length
        for _ in range(offset_length-cur_token_index):
            sub_labels.append('O')

        if sub_word_ids[0] is None:
            sub_labels[0] = self.config.start_label

        if sub_word_ids[offset_length-1] is None:
            sub_labels[offset_length-1] = self.config.end_label

        if len(sub_labels)!= offset_length:
            logger.error(f"{len(sub_labels)} vs {offset_length}")
            for i in one_ins:
                logger.error(f"{i}")
            raise PermissionError

        if not self.config.clean_droped_entity:
            entities_info = one_ins[self.config.entities_info]
        return sub_labels, entities_info



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hjson
import pickle as pkl
import os
import json
import pandas as pd
import torch
from typing import Union, Dict, Any, List
from dlk.data.postprocessors import postprocessor_register, postprocessor_config_register, IPostProcessor, IPostProcessorConfig
from dlk.utils.logger import Logger
from dlk.utils.vocab import Vocabulary
from dlk.utils.io import open
import torchmetrics

logger = Logger.get_logger()


@postprocessor_config_register('txt_reg')
class TxtRegPostProcessorConfig(IPostProcessorConfig):
    default_config = {
        "_name": "txt_reg",
        "config": {
            "input_map": {
                "logits": "logits",
                "values": "values",
                "_index": "_index",
            },
            "origin_input_map": {
                "sentence": "sentence",
                "sentence_a": "sentence_a", # for pair
                "sentence_b": "sentence_b",
                "uuid": "uuid"
            },
            "data_type": "single", # single or pair
            "save_root_path": ".",  # save data root dir
            "save_path": {
                "valid": "valid",  # relative dir for valid stage
                "test": "test",    # relative dir for test stage
            },
            "log_reg": False, # whether logistic regression
            "start_save_step": 0,  # -1 means the last
            "start_save_epoch": -1,
        }
    }
    """Config for TxtRegPostProcessor

    Config Example:
    """

    def __init__(self, config: Dict):
        super(TxtRegPostProcessorConfig, self).__init__(config)

        self.data_type = self.config['data_type']
        assert self.data_type in {'single', 'pair'}
        if self.data_type == 'pair':
            self.sentence_a = self.origin_input_map['sentence_a']
            self.sentence_b = self.origin_input_map['sentence_b']
        else:
            self.sentence = self.origin_input_map['sentence']
        self.uuid = self.origin_input_map['uuid']
        self.log_reg = self.config['log_reg']

        self.value = self.input_map['values']
        self.logits = self.input_map['logits']
        self._index = self.input_map['_index']
        self.save_path = self.config['save_path']
        self.save_root_path = self.config['save_root_path']
        self.start_save_epoch = self.config['start_save_epoch']
        self.start_save_step = self.config['start_save_step']
        self.post_check(self.config, used=[
            "input_map",
            "origin_input_map",
            "save_root_path",
            "save_path",
            "data_type",
            "start_save_step",
            "start_save_epoch",
            "log_reg",
        ])


@postprocessor_register('txt_reg')
class TxtRegPostProcessor(IPostProcessor):
    """text regression postprocess"""
    def __init__(self, config: TxtRegPostProcessorConfig):
        super(TxtRegPostProcessor, self).__init__(config)
        self.config = config

    def do_predict(self, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict)->List:
        """Process the model predict to human readable format

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns: 
            all predicts
        """
        results = []
        for outputs in list_batch_outputs:
            logits = outputs[self.config.logits].detach()
            if self.config.log_reg:
                logits = torch.sigmoid(logits)
            assert len(logits.shape) == 2
            # predict_indexes = list(torch.argmax(logits, 1))
            indexes = list(outputs[self.config._index])

            if self.config.value in outputs:
                values = outputs[self.config.value]
            else:
                values = [0.0] * len(indexes)
            for i, (one_logits, index, value) in enumerate(zip(logits, indexes, values)):
                one_ins = {}
                one_origin = origin_data.iloc[int(index)]
                if self.config.data_type == 'single':
                    sentence = one_origin[self.config.sentence]
                    one_ins['sentence'] = sentence
                else:
                    sentence_a = one_origin[self.config.sentence_a]
                    one_ins['sentence_a'] = sentence_a
                    sentence_b = one_origin[self.config.sentence_b]
                    one_ins['sentence_b'] = sentence_b
                    
                uuid = one_origin[self.config.uuid]
                one_ins['uuid'] = uuid
                one_ins['values'] = [float(value)]
                one_ins['predict_values'] = [float(one_logits)]
                one_ins['predict_extend_return'] = self.gather_predict_extend_data(outputs, i, self.config.predict_extend_return)
                results.append(one_ins)
        return results

    def do_calc_metrics(self, predicts: List, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict)->Dict:
        """calc the scores use the predicts or list_batch_outputs

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns: 
            the named scores

        """
        return {}

    def do_save(self, predicts: List, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict, save_condition: bool=False):
        """save the predict when save_condition==True

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            save_condition: True for save, False for depend on rt_config

        Returns: 
            None
        """

        if self.config.start_save_epoch == -1 or self.config.start_save_step == -1:
            self.config.start_save_step = rt_config.get('total_steps', 0) - 1
            self.config.start_save_epoch = rt_config.get('total_epochs', 0) - 1
        if not save_condition and (rt_config['current_step']>=self.config.start_save_step or rt_config['current_epoch']>=self.config.start_save_epoch):
            save_condition = True
        if save_condition:
            save_path = os.path.join(self.config.save_root_path, self.config.save_path.get(stage, ''))
            if "current_step" in rt_config:
                save_file = os.path.join(save_path, f"step_{str(rt_config['current_step'])}_predict.json")
            else:
                save_file = os.path.join(save_path, 'predict.json')
            logger.info(f"Save the {stage} predict data at {save_file}")
            with open(save_file, 'w') as f:
                json.dump(predicts, f, indent=4, ensure_ascii=False)




# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Iterable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from dlk.utils.logger import Logger
import pandas as pd

logger = Logger.get_logger()

@subprocessor_config_register('token_gather')
class TokenGatherConfig(BaseConfig):
    default_config = {
        "_name": "token_gather",
        "config": {
            "train": {
                "data_set": {                   # for different stage, this processor will process different part of data
                    "train": ["train", "valid", 'test']
                },
                "gather_columns": "*@*", # List of columns, if one element of the list is dict, you can set more. Every cell must be sigle token or list of tokens or set of tokens
                # gather_columns": ['tokens']
                # gather_columns": ['tokens', {"column": "entities_info", "trace": 'labels'}] 
                # the trace only trace the dict, if list is in trace path, will add the trace to every elements in the list. for example: {"entities_info": [{'start': 1， 'end': 2, labels: ['Label1']}, ..]}, the trace to labels is 'entities_info.labels'
                "deliver": "*@*", # output Vocabulary object (the Vocabulary of labels) name.
                "ignore": "", # ignore the token, the id of this token will be -1
                "update": None, # null or another Vocabulary object to update
                "unk": "[UNK]",
                "pad": "[PAD]",
                "min_freq": 1,
                "most_common": -1, # 1 for all
            }
        }
    }
    """Config for TokenGather

    Config Example:
        default_config
    """

    def __init__(self, stage: str, config: Dict):
        super(TokenGatherConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.ignore = self.config['ignore']
        self.gather_columns = self.config["gather_columns"]
        self.deliver = self.config["deliver"]
        if self.data_set and (not self.deliver):
            raise ValueError("The 'deliver' value must not be null.")
        self.update = self.config['update']
        self.unk = self.config['unk']
        self.pad = self.config['pad']
        self.min_freq = self.config['min_freq']
        self.most_common = self.config['most_common']
        self.post_check(self.config, used=[
            "data_set",
            "gather_columns",
            "deliver",
            "ignore",
            "update",
            "unk",
            "pad",
            "min_freq",
            "most_common",
        ])

@subprocessor_register('token_gather')
class TokenGather(ISubProcessor):
    """gather all tokens from the 'gather_columns' and deliver a vocab named 'token_vocab'
    """
    def __init__(self, stage: str, config: TokenGatherConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'token_gather' at stage {self.stage}")
            return
        self.update = config.update

    def get_elements_from_series_by_trace(self, data: pd.Series, trace: str)->List:
        """get the datas from data[trace_path]
        >>> for example:
        >>> data[0] = {'entities_info': [{'start': 0, 'end': 1, 'labels': ['Label1']}]} // data is a series, and every element is as data[0]
        >>> trace = 'entities_info.labels'
        >>> return_result = [['Label1']]

        Args:
            data: origin data series
            trace: get data element trace

        Returns: 
            the data in the tail of traces

        """

        def get_elements_from_iter_by_trace(iter: Iterable, cur_trace_list: List):
            if not cur_trace_list:
                return iter
            if isinstance(iter, dict):
                return get_elements_from_iter_by_trace(iter[cur_trace_list[0]], cur_trace_list[1:])
            if isinstance(iter, list) or  isinstance(iter, tuple):
                return [get_elements_from_iter_by_trace(sub_iter, cur_trace_list) for sub_iter in iter]
            raise PermissionError(f"The trace path is only support type list and dict, but you provide {type(iter)}")
            
        return [get_elements_from_iter_by_trace(one, trace.split('.')) for one in data]

    def process(self, data: Dict)->Dict:
        """TokenGather entry

        Args:
            data: 
            >>> {
            >>>     "data": {"train": ...},
            >>>     "tokenizer": ..
            >>> }

        Returns: 
            data[self.config.deliver] = Vocabulary()(which gathered_token)

        """
        if not self.data_set:
            return data
        if self.update:
            self.vocab = data[self.update]
        else:
            self.vocab = Vocabulary(do_strip=True, unknown=self.config.unk, ignore=self.config.ignore, pad=self.config.pad)
        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip gather tokens from it.')
                continue
            data_set = data['data'][data_set_name]
            for column in self.config.gather_columns:
                if isinstance(column, str):
                    self.vocab.auto_update(data_set[column])
                elif isinstance(column, dict):
                    self.vocab.auto_update(self.get_elements_from_series_by_trace(data_set[column['column']], trace=column['trace']))
                else:
                    raise PermissionError(f'The gather column currently is only support str or dict.')
        self.vocab.filter_rare(self.config.min_freq, self.config.most_common)
        logger.info(f"The Vocab Num is {self.vocab.word_num}")
        data[self.config.deliver] = self.vocab.__dict__
        return data



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
import math
from . import scheduler_register, scheduler_config_register, BaseScheduler, BaseSchedulerConfig
from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts
import torch.optim as optim
from dlk.utils.logger import Logger
logger = Logger.get_logger()


@scheduler_config_register("cosine_restart")
class CosineRestartScheduleConfig(BaseSchedulerConfig):
    default_config = {
        "config": {
            "first_restart_step": -1, # you can just set first_restart_step or first_restart_epoch
            "first_restart_epoch": -1,
            "mult_fact": 1,
            },
        "_name": "cosine_restart",
    }
    """Config for CosineRestartSchedule

    Config Example: see default_config
    """
    def __init__(self, config: Dict):
        super(CosineRestartScheduleConfig, self).__init__(config)
        config = config['config']
        self.first_restart_epoch = config["first_restart_epoch"]
        self.first_restart_step = config["first_restart_step"]
        assert self.first_restart_epoch != -1 or self.first_restart_step != -1, "You must provide one of them"
        assert self.first_restart_epoch == -1 or self.first_restart_step == -1, "You must provide one of them"
        self.mult_fact = config['mult_fact']
        self.post_check(config, used=[
            "mult_fact",
            "first_restart_epoch",
            "first_restart_step",
        ])


@scheduler_register("cosine_restart")
class CosineRestartSchedule(BaseScheduler):
    """CosineRestartSchedule"""
    def __init__(self, optimizer: optim.Optimizer, config: CosineRestartScheduleConfig):
        super(CosineRestartSchedule, self).__init__()
        self.config = config
        self.optimizer = optimizer

    def get_scheduler(self)->CosineAnnealingWarmRestarts:
        """return the initialized linear wramup then cos decay scheduler

        Returns: 
            Schedule

        """
        epoch_training_steps = self.config.epoch_training_steps
        num_training_steps = self.config.num_training_steps
        if self.config.first_restart_step == -1:
            self.config.first_restart_step = self.config.first_restart_epoch * epoch_training_steps
        logger.warning(f"The calculated Total Traning Num is {num_training_steps}, the Restart Steps is {self.config.first_restart_step}. Please check it carefully.")

        return CosineAnnealingWarmRestarts(self.optimizer, self.config.first_restart_step, self.config.mult_fact, last_epoch=-1)



# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http:# www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from typing import Dict, List, Set, Callable
from dlk.core.base_module import SimpleModule, BaseModuleConfig
from dlk.utils.config import BaseConfig, ConfigTool
from . import decoder_register, decoder_config_register
import torch.nn as nn

@decoder_config_register("multi_decoder")
class MultiDecoderConfig(BaseModuleConfig):
    default_config = {
        "config": {
        },
        "_name": "multi_decoder",
    }
    """Config for MultiDecoder 
    """
    def __init__(self, config: Dict):
        super(MultiDecoderConfig, self).__init__(config)
        self.decode_configs = {}
        for decode in config:
            if decode in {"config", "_name", "_base"}:
                continue
            module_class, module_config = ConfigTool.get_leaf_module(decoder_register, decoder_config_register, "decoder", config[decode])
            self.decode_configs[decode] = {
                "decode_class": module_class,
                "decode_config": module_config,
            }


@decoder_register("multi_decoder")
class MultiDecoder(SimpleModule):
    """multi_decoder a x A x b
    """
    def __init__(self, config: MultiDecoderConfig):
        super(MultiDecoder, self).__init__(config)
        self._provide_keys = {'logits'}
        self._required_keys = {'embedding'}
        self._provided_keys = set()
        decode_configs = config.decode_configs
        self.decoders = nn.ModuleDict({
            decode_name: decode_configs[decode_name]['decode_class'](decode_configs[decode_name]['decode_config'])
            for decode_name in decode_configs
        })

        self.config = config

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns: 
            None

        """
        for decode_name in self.decoders:
            self.decoders[decode_name].init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        for decode_name in self.config.decode_configs:
            inputs = self.decoders[decode_name](inputs)
        return inputs



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
import torch.nn as nn
from . import loss_register, loss_config_register
from packaging import version
import torch.nn as nn
from dlk.utils.logger import Logger
from dlk.core.base_module import BaseModuleConfig
import torch

logger = Logger.get_logger()

@loss_config_register("bce")
class BCEWithLogitsLossConfig(BaseModuleConfig):
    default_config ={
        "config": {
            "pred_truth_pair": [], # len(.) == 2, the 1st is the pred_name, 2nd is truth_name in __call__ inputs
            "masked_select": None, # if provide, only select the masked(=1) data
            "schedule": [1],
            "scale": [1], # scale the loss for every schedule stage
            "log_map": {
                "loss": "loss"
            }
            # "schdeule": [0.3, 1.0], # can be a list or str
            # "scale": "[0.5, 1]",
            },
        "_name": "bce",
    }
    """Config for BCEWithLogitsLoss

    Config Example: default_config
    """
    def __init__(self, config: Dict):
        super(BCEWithLogitsLossConfig, self).__init__(config)
        config = config['config']

        self.scale = config['scale']
        self.schedule = config['schedule']

        if isinstance(self.scale, str):
            self.scale = eval(self.scale)
        if isinstance(self.schedule, str):
            self.schedule = eval(self.schedule)

        if not isinstance(self.scale, list):
            assert isinstance(float(self.scale), float)
            self.scale = [self.scale]
        if not isinstance(self.schedule, list):
            assert isinstance(float(self.schedule), float)
            self.schedule = [self.schedule]
        assert len(self.schedule) == len(self.scale)
        assert self.schedule[-1] - 1 < 0.00001

        self.pred_truth_pair = config['pred_truth_pair']
        self.log_map = config['log_map']
        if isinstance(self.log_map, str):
            self.log_map = {"loss": self.log_map}
        if not self.pred_truth_pair:
            raise PermissionError(f"You must provide the pred_truth_pair for loss.")
        self.masked_select = config['masked_select']
        self.post_check(config, used=[
            "pred_truth_pair",
            "masked_select",
            "schedule",
            "scale",
            "log_map",
        ])


@loss_register("bce")
class BCEWithLogitsLoss(nn.Module):
    """binary crossentropy for bi-class classification
    """
    def __init__(self, config: BCEWithLogitsLossConfig):
        super(BCEWithLogitsLoss, self).__init__()
        self.config = config
        self.bce = nn.BCEWithLogitsLoss(reduction='mean')

    def update_config(self, rt_config: Dict):
        """callback for imodel to update the total steps and epochs

        when init the loss module, the total step and epoch is not known, when all data ready, the imodel update the value for loss module

        Args:
            rt_config: { "total_steps": self.num_training_steps, "total_epochs": self.num_training_epochs}

        Returns: 
            None

        """
        self.current_stage = 0
        self.config.schedule = [rt_config['total_steps']*i for i in self.config.schedule]

    def calc(self, result, inputs, rt_config):
        """calc the loss the predict is from result, the ground truth is from inputs

        Args:
            result: the model predict dict
            inputs: the all inputs for model
            rt_config: provide the current training status 
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns: 
            loss

        """
        if rt_config['current_step']>self.config.schedule[self.current_stage]:
            self.current_stage += 1
        scale = self.config.scale[self.current_stage]
        pred_name, truth_name = self.config.pred_truth_pair
        pred = result[pred_name]
        target = inputs[truth_name]
        batch_size = target.shape[0]
        if self.config.masked_select is not None:
            pred = torch.masked_select(pred, inputs[self.config.masked_select])
            target = torch.masked_select(target, inputs[self.config.masked_select])
        loss = self.bce(torch.sigmoid(pred), target) * scale
        return loss, {self.config.log_map['loss']: loss}

    def __call__(self, result, inputs, rt_config):
        """same as self.calc
        """
        return self.calc(result, inputs, rt_config)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
import pandas as pd
from dlk.utils.logger import Logger

logger = Logger.get_logger()

@subprocessor_config_register('basic_data_loader')
class BasicDataLoaderConfig(BaseConfig):
    default_config = {
        "_name": "basic_data_loader@seq_lab",
        "config": {
            "train": "*@*",
            "extend_train": "train",
            "predict": "*@*",
            "online": "*@*",
        }
    }
    """Config for BasicDataLoader

    Config Example:
        default_config
    """
    def __init__(self, stage, config: Dict):

        super(BasicDataLoaderConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.output_map = self.config.get('output_map', {})
        self.input_map = self.config.get('input_map', {})
        self.post_check(self.config, used=[
            "data_set",
            "input_map",
            "output_map",
        ])


@subprocessor_register('basic_data_loader')
class BasicDataLoader(ISubProcessor):
    """
    Loader the data from dict and generator DataFrame
    """

    def __init__(self, stage: str, config: BasicDataLoaderConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'basic_data_loader' at stage {self.stage}")
            return

    def process(self, data: Dict)->Dict:
        """example for sequence labeling loader

        Args:
            data: 
                >>> input data
                >>> {
                >>>     "train": list of json format train data
                >>> }
                >>> one_ins example:
                >>> {
                >>>     "uuid": '**-**-**-**'
                >>>     "sentence": "I have an apple",
                >>>     "labels": [
                >>>                 {
                >>>                     "end": 15,
                >>>                     "start": 10,
                >>>                     "labels": [
                >>>                         "Fruit"
                >>>                     ]
                >>>                 },
                >>>                 ...,
                >>>             ]
                >>>         },
                >>>     ],
                >>> }
        Returns: 
            data + loaded_data
        """

        if not self.data_set:
            return data

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do basic_data_loader on it.')
                continue
            data_set = data['data'][data_set_name]

            input_data = {key: [] for key in self.config.input_map.keys()}

            for one_ins in data_set:
                try:
                    for key in self.config.input_map:
                        input_data[key].append(one_ins[self.config.input_map[key]])
                except:
                    raise PermissionError(f"You must provide the data as requests, we need {self.config.input_map.keys()}, or you can provide the input_map to map the origin data to this format")
            output_data = {self.config.output_map[key]: input_data[key] for key in self.config.output_map.keys()}
            data_df = pd.DataFrame(data= output_data)
            data['data'][data_set_name] = data_df

        return data



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tokenizers.processors import TemplateProcessing
from tokenizers.normalizers import Lowercase, NFD, NFC, StripAccents, Strip
from tokenizers.pre_tokenizers import WhitespaceSplit, ByteLevel, Whitespace, BertPreTokenizer
from tokenizers import Tokenizer
from dlk.utils.logger import Logger

logger = Logger.get_logger()


class TokenizerPostprocessorFactory(object):
    """docstring for TokenizerPostprocessorFactory"""
    def __init__(self, tokenizer: Tokenizer):
        self.tokenizer = tokenizer

    @property
    def bert(self):
        """bert postprocess

        Returns: 
            bert postprocess

        """
        vocab = self.tokenizer.get_vocab()
        try:
            cls_id = vocab['[CLS]']
            sep_id = vocab['[SEP]']
        except Exception as e:
            # logger.error(f"`[CLS]` or `[SEP]` is not a token in this tokenizer.", )
            logger.exception(f"`[CLS]` or `[SEP]` is not a token in this tokenizer.")
            raise e
            # raise PermissionError(f"`[CLS]` or `[SEP]` is not a token in this tokenizer.")

        def _wrap():
            return TemplateProcessing(
                single="[CLS] $A [SEP]",
                pair="[CLS] $A [SEP] $B:1 [SEP]:1",
                special_tokens=[
                    ("[CLS]", cls_id),
                    ("[SEP]", sep_id),
                ],
            )
        return _wrap


    def get(self, name):
        """get postprocess by name 

        Returns: 
            postprocess

        """
        return self.__getattribute__(name)



class PreTokenizerFactory(object):
    """PreTokenizerFactory"""
    def __init__(self, tokenizer: Tokenizer):
        self.tokenizer = tokenizer

    @property
    def bytelevel(self):
        """byte level pre_tokenizer

        Returns: 
            ByteLevel

        """
        return ByteLevel

    @property
    def bert(self):
        """bert pre_tokenizer

        Returns: 
            BertPreTokenizer

        """
        return BertPreTokenizer

    @property
    def whitespace(self):
        """whitespace pre_tokenizer

        Returns: 
            Whitespace

        """
        return Whitespace

    @property
    def whitespacesplit(self):
        """whitespacesplit pre_tokenizer

        Returns: 
            WhitespaceSplit

        """
        return WhitespaceSplit

    def get(self, name):
        """get pretokenizer by name 

        Returns: 
            postprocess

        """
        return self.__getattribute__(name)


class TokenizerNormalizerFactory(object):
    """TokenizerNormalizerFactory"""
    def __init__(self, tokenizer: Tokenizer):
        self.tokenizer = tokenizer

    @property
    def lowercase(self):
        """do lowercase normalizers

        Returns: 
            Lowercase

        """
        return Lowercase

    @property
    def nfd(self):
        """do nfd normalizers

        Returns: 
            NFD

        """
        return NFD

    @property
    def nfc(self):
        """do nfc normalizers

        Returns: 
            NFC

        """
        return NFC

    @property
    def strip_accents(self):
        """do strip normalizers

        Returns: 
            StripAccents

        """
        return StripAccents

    @property
    def strip(self):
        """do strip normalizers

        Returns: 
            StripAccents

        """
        return Strip

    def get(self, name):
        """get normalizers by name 

        Returns: 
            Normalizer

        """
        return self.__getattribute__(name)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from loguru import logger as logging
from loguru._logger import Logger as LoggerClass
from typing import Set
import sys
import os


class Logger(object):
    """docstring for logger"""
    global_logger: LoggerClass = None
    global_log_file: Set[str] = set()
    log_name: str = "dlk"
    warning_file = True

    level_map = {
        "debug": "DEBUG",
        "info": "INFO",
        "warning": "WARNING",
        "error": "ERROR",
    }

    def __init__(self, log_file: str='', base_dir: str='logs', log_level: str='debug', log_name="dlk"):
        super(Logger, self).__init__()
        self.log_file = log_file
        self.base_dir = base_dir
        Logger.log_name = log_name
        if self.base_dir and not os.path.isdir(self.base_dir):
            os.mkdir(self.base_dir)

        self.init_global_logger(log_level=log_level, log_name=Logger.log_name, reinit=True)
        self.init_file_logger(log_file, base_dir, log_level=log_level)

    @staticmethod
    def get_logger()->LoggerClass:
        """return the 'dlk' logger if initialized otherwise init and return it

        Returns: 
            Logger.global_logger

        """
        if Logger.global_logger is None:
            Logger.init_global_logger()
        if not Logger.global_log_file and Logger.warning_file:
            Logger.global_logger.warning("You didn't add the logger file, only stdout(stderr) is working.")
            Logger.warning_file = False
        return Logger.global_logger

    @staticmethod
    def init_file_logger(log_file, base_dir='logs', log_level: str='debug'):
        """init(if there is not one) or change(if there already is one) the log file

        Args:
            log_file: log file path
            base_dir: real log path is '$base_dir/$log_file'
            log_level: 'debug', 'info', etc.

        Returns: 
            None
        """
        if "LOCAL_RANK" in os.environ and os.environ["LOCAL_RANK"] not in [-1, 0]:
            return 
        if "GLOBAL_RANK" in os.environ and os.environ["GLOBAL_RANK"] != 0:
            return 
        if log_file:
            log_file = os.path.join(base_dir, log_file)
        if Logger.global_log_file:
            Logger.global_logger.warning(f"Exists a file handler at '{'; '.join(Logger.global_log_file)}'")
        Logger.global_log_file.add(log_file)
        Logger.global_logger.add(log_file, rotation="10 MB", format="{time:MM/DD/YYYY HH:mm:ss} - {level:<8} - "+Logger.log_name+" - {message}")

    @staticmethod
    def init_global_logger(log_level: str='debug', log_name: str=None, reinit: bool=False):
        """init the global_logger

        Args:
            log_level: you can change this to logger to different level
            log_name: change this is not suggested 
            reinit: if set true, will force reinit

        Returns: 
            None

        """
        if log_name and (log_name != Logger.log_name):
            Logger.log_name = log_name
        if reinit or not Logger.global_logger:
            Logger.global_logger = logging
            Logger.global_logger.remove()

            if ("LOCAL_RANK" in os.environ and os.environ["LOCAL_RANK"] not in [-1, 0]) or ("GLOBAL_RANK" in os.environ and os.environ["GLOBAL_RANK"] != 0):
                Logger.global_logger.add(sys.stdout, level=Logger.level_map["error"], format="<level>{time:MM/DD/YYYY HH:mm:ss} - {level:<8}</level> - <cyan>"+Logger.log_name+"</cyan> - <level>{message}</level>")
            else:
                Logger.global_logger.add(sys.stdout, level=Logger.level_map[log_level], format="<level>{time:MM/DD/YYYY HH:mm:ss} - {level:<8}</level> - <cyan>"+Logger.log_name+"</cyan> - <level>{message}</level>")
            # Logger.global_logger.level("INFO",  color="<g>")




# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""optimizers"""
import importlib
import os
from typing import Callable, Dict, Tuple, Any
from dlk.utils.register import Register
import torch.optim as optim
from dlk.utils.logger import Logger
import torch
import re


logger = Logger.get_logger()
optimizer_config_register = Register("Optimizer config register.")
optimizer_register = Register("Optimizer register.")


class BaseOptimizer(object):

    def get_optimizer(self)->optim.Optimizer:
        """return the initialized optimizer

        Returns: 
            Optimizer

        """
        raise NotADirectoryError

    def init_optimizer(self, optimizer: optim.Optimizer, model: torch.nn.Module, config: Dict):
        """init the optimizer for paras in model, and the group is decided by config

        Args:
            optimizer: adamw, sgd, etc.
            model: pytorch model
            config: which decided the para group, lr, etc.

        Returns: 
            the initialized optimizer

        """
        optimizer_special_groups = config.pop('optimizer_special_groups', {})
        params = []
        all_named_parameters = list(model.named_parameters())
        total_all_named_parameters = len(all_named_parameters)
        logger.info(f"All Named Params Num is {len(all_named_parameters)}")
        has_grouped_params = set()
        for special_group_name in optimizer_special_groups.get('order', []):

            group_config = optimizer_special_groups[special_group_name]['config']
            group_patterns = optimizer_special_groups[special_group_name]['pattern']
            # convert to regex
            combine_patterns = []
            for pattern in group_patterns:
                combine_patterns.append(f"({pattern})")
            cc_patterns = re.compile("|".join(combine_patterns))
            group_param = []
            for n, p  in all_named_parameters:
                # logger.info(f"Param name {n}")
                if n in has_grouped_params:
                    continue
                if cc_patterns.search(n): # use regex
                    has_grouped_params.add(n)
                    group_param.append(p)
            group_config['params'] = group_param
            group_config['name'] = special_group_name
            params.append(group_config)

        reserve_params = [p for n, p in all_named_parameters if not n in has_grouped_params]
        params.append({"params": reserve_params, "name": config.pop('name')})
        logger.info(f"Param Group Nums {len(params)}")
        total_param = 0
        for group in params:
            total_param = total_param + len(group['params'])
        assert total_param == total_all_named_parameters

        return optimizer(params=params, **config)

    def __call__(self):
        """the same as self.get_optimizer()
        """
        return self.get_optimizer()


def import_optimizers(optimizers_dir, namespace):
    for file in os.listdir(optimizers_dir):
        path = os.path.join(optimizers_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            optimizer_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + optimizer_name)


# automatically import any Python files in the optimizers directory
optimizers_dir = os.path.dirname(__file__)
import_optimizers(optimizers_dir, "dlk.core.optimizers")



# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
from dlk.utils.logger import Logger
import numpy as np
import pandas as pd
import os

logger = Logger.get_logger()

@subprocessor_config_register('piece_rerank_relabel')
class PieceRerankRelabelConfig(BaseConfig):
    default_config = {
        "_name": "piece_rerank_relabel",
        "config": {
            "train":{
                "input_map": {  
                     "word_ids": "word_ids",
                     "offsets": "offsets",
                     "rank_info": "rank_info",
                },
                "data_set": {                   # for different stage, this processor will process different part of data
                     "train": ['train', 'valid', 'test'],
                     "predict": ['predict'],
                },
                "output_map": {
                     "label_ids": "label_ids",
                },
                "mask_fill": -100,
            },
            "extend_train": "train"
        }
    }
    """Config for PieceRerankRelabel
    Config Example: 
        default_config
    """
    def __init__(self, stage, config: Dict):

        super(PieceRerankRelabelConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.word_ids = self.config['input_map']['word_ids']
        self.offsets = self.config['input_map']['offsets']
        self.mask_fill = self.config['mask_fill']
        self.rank_info = self.config['input_map']['rank_info']
        self.output_labels = self.config['output_map']['label_ids']
        self.post_check(self.config, used=[
            "input_map",
            "data_set",
            "output_map",
            "mask_fill"
        ])


@subprocessor_register('piece_rerank_relabel')
class PieceRerankRelabel(ISubProcessor):
    """
    Relabel the piece rank construct matrix
    """

    def __init__(self, stage: str, config: PieceRerankRelabelConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'piece_rerank_relabel' at stage {self.stage}")
            return

    def process(self, data: Dict)->Dict:
        """PieceRerankRelabel Entry

        Args:
            data: Dict

        Returns: 
            
            relabeled data

        """

        if not self.data_set:
            return data

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do piece_rerank_relabel on it.')
                continue
            data_set = data['data'][data_set_name]
            if os.environ.get('DISABLE_PANDAS_PARALLEL', 'false') != 'false':
                data_set[self.config.output_labels] = data_set.parallel_apply(self.relabel, axis=1)
            else:
                data_set[self.config.output_labels] = data_set.apply(self.relabel, axis=1)

        return data

    def relabel(self, one_ins: pd.Series):
        """make token label, if use the first piece label please use the 'piece_rerank_firstpiece_relabel'

        Args:
            one_ins: include sentence, rank_info

        Returns: 
            label_matrix for real rank
        """
        word_ids: List = one_ins[self.config.word_ids]
        rank_info = one_ins[self.config.rank_info]
        seq_len = len(word_ids)
        label_matrix = np.full((seq_len, seq_len), 0, dtype=np.int8)
        if not word_ids[0]:
            label_matrix[0] = self.config.mask_fill
            label_matrix[:, 0] = self.config.mask_fill
        if not word_ids[-1]:
            label_matrix[-1] = self.config.mask_fill
            label_matrix[:, -1] = self.config.mask_fill
        # {
        #     "uuid": "uuid",
        #     "pretokenized_words": ['[PAD]', "BCD", "E", "A"],
        #     "rank_info": [0, 3, 1, 2]
        # }
        # ['[CLS]', '[PAD]', 'BC', '##D', 'E', 'A', '[SEP]']
        # [(0, 0), (0, 5), (0, 2), (2, 3), (0, 1), (0, 1), (0, 0)]
        # [None, 0, 1, 1, 2, 3, None]
        # [101, 0, 3823, 2137, 142, 138, 102]
        assert rank_info[0] == 0
        pre_position = None
        for index in rank_info:
            cur_position = word_ids.index(index)
            cur_word_id = word_ids[cur_position]
            if pre_position:
                label_matrix[pre_position][cur_position] = 1
            while cur_position+2 < seq_len and word_ids[cur_position+1] == cur_word_id:
                label_matrix[cur_position][cur_position+1] = 1
                cur_position = cur_position + 1
            pre_position = cur_position
        start_position = word_ids.index(0)
        label_matrix[pre_position][start_position] = 1
        return label_matrix



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
import torch
from typing import Dict, List, Set, Callable
from dlk.core.base_module import SimpleModule, BaseModuleConfig
from . import encoder_register, encoder_config_register
from dlk.core.modules import module_config_register, module_register
from dlk.utils.logger import Logger
logger = Logger.get_logger()

@encoder_config_register("lstm")
class LSTMConfig(BaseModuleConfig):
    """Config for LSTM

    Config Example:
        >>> {
        >>>     module: {
        >>>         _base: "lstm",
        >>>     },
        >>>     config: {
        >>>         input_map: {},
        >>>         output_map: {},
        >>>         input_size: *@*,
        >>>         output_size: "*@*",
        >>>         num_layers: 1,
        >>>         dropout: "*@*", // dropout between layers
        >>>     },
        >>>     _link: {
        >>>         config.input_size: [module.config.input_size],
        >>>         config.output_size: [module.config.output_size],
        >>>         config.dropout: [module.config.dropout],
        >>>     },
        >>>     _name: "lstm",
        >>> }
    """

    def __init__(self, config: Dict):
        super(LSTMConfig, self).__init__(config)
        self.lstm_config = config["module"]
        assert self.lstm_config['_name'] == "lstm"
        self.post_check(config['config'], used=[
            "input_size",
            "output_size",
            "num_layers",
            "return_logits",
            "dropout",
        ])


@encoder_register("lstm")
class LSTM(SimpleModule):
    """Wrap for torch.nn.LSTM
    """
    def __init__(self, config: LSTMConfig):
        super(LSTM, self).__init__(config)
        self._provide_keys = {'embedding'}
        self._required_keys = {'embedding', 'attention_mask'}
        self._provided_keys = set()
        self.config = config
        self.lstm = module_register.get('lstm')(module_config_register.get('lstm')(config.lstm_config))

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns: 
            None

        """
        self.lstm.init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """All step do this

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        inputs[self.get_output_name('embedding')] = self.lstm(inputs[self.get_input_name('embedding')], inputs[self.get_input_name('attention_mask')])
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([inputs[self.get_output_name('embedding')]]))
        return inputs



# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http:// www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle as pkl
import json
from typing import Dict, List, Optional, Tuple, Union
import os
import numpy as np
import pandas as pd
import torch
from typing import Dict
from dlk.data.postprocessors import postprocessor_register, postprocessor_config_register, IPostProcessor, IPostProcessorConfig
from dlk.utils.logger import Logger
from dlk.utils.vocab import Vocabulary
from dlk.utils.io import open
from tokenizers import Tokenizer
import torchmetrics
logger = Logger.get_logger()


@postprocessor_config_register('span_cls')
class SpanClsPostProcessorConfig(IPostProcessorConfig):
    default_config = {
            "_name": "span_cls",
            "config": {
                "meta": "*@*",
                "ignore_position": False, # calc the metrics, whether ignore the ground_truth and predict position info.( if set to true, only focus on the entity content not position.)
                "ignore_char": " ", # if the entity begin or end with this char, will ignore these char
                # "ignore_char": " ()[]-.,:", # if the entity begin or end with this char, will ignore these char
                "meta_data": {
                    "label_vocab": 'label_vocab',
                    "tokenizer": "tokenizer",
                },
                "input_map": {
                    "logits": "logits",
                    "_index": "_index",
                },
                "origin_input_map": {
                    "uuid": "uuid",
                    "sentence": "sentence",
                    "input_ids": "input_ids",
                    "entities_info": "entities_info",
                    "offsets": "offsets",
                    "special_tokens_mask": "special_tokens_mask",
                    "word_ids": "word_ids",
                    "label_ids": "label_ids",
                },
                "save_root_path": ".",  # save data root dir
                "save_path": {
                    "valid": "valid",  # relative dir for valid stage
                    "test": "test",    # relative dir for test stage
                },
                "start_save_step": 0,  # -1 means the last
                "start_save_epoch": -1,
                "ignore_labels": ['O', 'X', 'S', "E"], # Out, Out, Start, End
            }
        }
    """Config for SpanClsPostProcessor

    Config Example:
    """

    def __init__(self, config: Dict):
        super(SpanClsPostProcessorConfig, self).__init__(config)
        self.ignore_labels = set(self.config['ignore_labels'])
        self.ignore_char = set(self.config['ignore_char'])
        self.ignore_position = self.config['ignore_position']

        self.sentence = self.origin_input_map['sentence']
        self.offsets = self.origin_input_map['offsets']
        self.entities_info = self.origin_input_map['entities_info']
        self.uuid = self.origin_input_map['uuid']
        self.word_ids = self.origin_input_map['word_ids']
        self.special_tokens_mask = self.origin_input_map['special_tokens_mask']
        self.input_ids = self.origin_input_map['input_ids']
        self.label_ids = self.origin_input_map['label_ids']

        self.logits = self.input_map['logits']
        self._index = self.input_map['_index']

        if isinstance(self.config['meta'], str):
            with open(self.config['meta'], 'rb') as f:
                meta = pkl.load(f)
        else:
            raise PermissionError("You must provide meta data(vocab & tokenizer) for ner postprocess.")

        vocab_trace_path = []
        vocab_trace_path_str = self.config['meta_data']['label_vocab']
        if vocab_trace_path_str and vocab_trace_path_str.strip()!='.':
            vocab_trace_path = vocab_trace_path_str.split('.')
        assert vocab_trace_path, "We need vocab and tokenizer all in meta, so you must provide the trace path from meta"
        self.label_vocab = meta
        for trace in vocab_trace_path:
            self.label_vocab = self.label_vocab[trace]
        self.label_vocab = Vocabulary.load(self.label_vocab)


        tokenizer_trace_path = []
        tokenizer_trace_path_str = self.config['meta_data']['tokenizer']
        if tokenizer_trace_path_str and tokenizer_trace_path_str.strip()!='.':
            tokenizer_trace_path = tokenizer_trace_path_str.split('.')
        assert tokenizer_trace_path, "We need vocab and tokenizer all in meta, so you must provide the trace path from meta"
        tokenizer_config = meta
        for trace in tokenizer_trace_path:
            tokenizer_config = tokenizer_config[trace]
        self.tokenizer = Tokenizer.from_str(tokenizer_config)

        self.save_path = self.config['save_path']
        self.save_root_path = self.config['save_root_path']
        self.start_save_epoch = self.config['start_save_epoch']
        self.start_save_step = self.config['start_save_step']

        self.post_check(self.config, used=[
            "meta",
            "meta_data",
            "input_map",
            "origin_input_map",
            "save_root_path",
            "save_path",
            "start_save_step",
            "start_save_epoch",
            "ignore_labels",
        ])


@postprocessor_register('span_cls')
class SpanClsPostProcessor(IPostProcessor):
    """PostProcess for sequence labeling task"""
    def __init__(self, config: SpanClsPostProcessorConfig):
        super(SpanClsPostProcessor, self).__init__(config)
        self.label_vocab = self.config.label_vocab
        self.tokenizer = self.config.tokenizer
        self.metric = torchmetrics.Accuracy()

    def _process4predict(self, predict_logits: torch.FloatTensor, index: int, origin_data: pd.DataFrame)->Dict:
        """gather the predict and origin text and ground_truth_entities_info for predict

        Args:
            predict: the predict span logits
            index: the data index in origin_data
            origin_data: the origin pd.DataFrame

        Returns: 
            >>> one_ins info 
            >>> {
            >>>     "sentence": "...",
            >>>     "uuid": "..",
            >>>     "entities_info": [".."],
            >>>     "predict_entities_info": [".."],
            >>> }

        """

        def _get_entity_info(sub_tokens_index: List, offset_mapping: List, word_ids: List, label: str)->Dict:
            """gather sub_tokens to get the start and end

            Args:
                sub_tokens_index: the entity tokens index list
                offset_mapping: every token offset in text
                word_ids: every token in the index of words
                label: predict label

            Returns: 
                entity_info

            """
            if not sub_tokens_index or not label:
                return {}
            start = offset_mapping[sub_tokens_index[0]][0]
            end = offset_mapping[sub_tokens_index[-1]][1]
            return {
                "start": start,
                "end": end,
                "labels": [label]
            }
        one_ins = {}
        origin_ins = origin_data.iloc[int(index)]
        one_ins["sentence"] = origin_ins[self.config.sentence]
        one_ins["uuid"] = origin_ins[self.config.uuid]
        one_ins["entities_info"] = origin_ins[self.config.entities_info]

        word_ids = origin_ins[self.config.word_ids]
        rel_token_len = len(word_ids)
        offset_mapping = origin_ins[self.config.offsets][:rel_token_len]

        predict_entities_info = []
        predict_label_ids = predict_logits.argmax(-1).cpu().numpy()
        for i in range(rel_token_len):
            for j in range(i, rel_token_len):
                if word_ids[i] is None or word_ids[j] is None:
                    continue
                predict_label_id = predict_label_ids[i][j]
                predict_label = self.config.label_vocab[predict_label_id]
                if predict_label == self.config.label_vocab.pad or predict_label == self.config.label_vocab.unknown:
                    continue
                else:
                    entity_info = _get_entity_info([i, j], offset_mapping, word_ids, predict_label)
                    if entity_info:
                        predict_entities_info.append(entity_info)
        one_ins['predict_entities_info'] = predict_entities_info
        return one_ins

    def do_predict(self, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict)->List:
        """Process the model predict to human readable format

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns: 
            all predicts

        """
        predicts = []
        if self.config.sentence not in origin_data:
            logger.error(f"{self.config.sentence} not in the origin data")
            raise PermissionError(f"{self.config.sentence} must be provided")
        if self.config.uuid not in origin_data:
            logger.error(f"{self.config.uuid} not in the origin data")
            raise PermissionError(f"{self.config.uuid} must be provided")

        predicts = []
        for outputs in list_batch_outputs:
            batch_logits = outputs[self.config.logits]
            # batch_special_tokens_mask = outputs[self.config.special_tokens_mask]

            indexes = list(outputs[self.config._index])

            for i, (predict, index) in enumerate(zip(batch_logits, indexes)):
                one_ins = self._process4predict(predict, index, origin_data)
                one_ins['predict_extend_return'] = self.gather_predict_extend_data(outputs, i, self.config.predict_extend_return)
                predicts.append(one_ins)
        return predicts

    def do_calc_metrics(self, predicts: List, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict)->Dict:
        """calc the scores use the predicts or list_batch_outputs

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns: 
            the named scores, recall, precision, f1

        """

        def _group_entities_info(entities_info: List[Dict], text: str)->Dict:
            """gather the same labeled entity to the same list

            Args:
                entities_info: 
                    >>> [
                    >>>     {
                    >>>         "start": start1,
                    >>>         "end": end1,
                    >>>         "labels": ["label_1"]
                    >>>     },
                    >>>     {
                    >>>         "start": start2,
                    >>>         "end": end2,
                    >>>         "labels": ["label_2"]
                    >>>     },....
                    >>> ]
                text: be labeled text

            Returns: 
                >>> { "label_1" [text[start1:end1]], "label_2": [text[start_2: end_2]]...}

            """
            info = {}
            for item in entities_info:
                label = item['labels'][0]
                if label not in info:
                    info[label] = []
                start_position, end_position = item['start'], item['end']
                while start_position < end_position:
                    if text[start_position] in self.config.ignore_char:
                        start_position += 1
                    else:
                        break
                while start_position < end_position:
                    if text[end_position-1] in self.config.ignore_char:
                        end_position -= 1
                    else:
                        break
                if start_position == end_position: # if the entity after remove ignore char be null, we set it to origin
                    start_position, end_position = item['start'], item['end']
                    
                if self.config.ignore_position:
                    info[label].append(text[item['start']: item['end']].strip())
                else:
                    info[label].append((start_position, end_position))
            return info

        def _calc_score(predict_list: List, ground_truth_list: List):
            """use predict_list and ground_truth_list to calc scores

            Args:
                predict_list: list of predict
                ground_truth_list: list of ground_truth

            Returns: 
                precision, recall, f1

            """
            category_tp = {}
            category_fp = {}
            category_fn = {}
            def _care_div(a, b):
                """return a/b or 0.0 if b == 0
                """
                if b==0:
                    return 0.0
                return a/b


            def _calc_num(_pred: List, _ground_truth: List):
                """calc tp, fn, fp

                Args:
                    pred: pred list
                    ground_truth: groud truth list

                Returns: 
                    tp, fn, fp

                """
                num_p = len(_pred)
                num_t = len(_ground_truth)
                truth = 0
                for p in _pred:
                    if p in _ground_truth:
                        truth += 1
                return truth, num_t-truth, num_p-truth

            for predict, ground_truth in zip(predict_list, ground_truth_list):
                keys = set(list(predict.keys())+list(ground_truth.keys()))
                for key in keys:
                    tp, fn, fp = _calc_num(predict.get(key, []), ground_truth.get(key, []))
                    # logger.info(f"{key} tp num {tp}, fn num {fn}, fp num {fp}")
                    category_tp[key] = category_tp.get(key, 0) + tp
                    category_fn[key] = category_fn.get(key, 0) + fn
                    category_fp[key] = category_fp.get(key, 0) + fp

            all_tp, all_fn, all_fp = 0, 0, 0
            for key in category_tp:
                tp, fn, fp = category_tp[key], category_fn[key], category_fp[key]
                all_tp += tp
                all_fn += fn
                all_fp += fp
                precision = _care_div(tp, tp+fp)
                recall = _care_div(tp, tp+fn)
                f1 = _care_div(2*precision*recall, precision+recall)
                logger.info(f"For entity 「{key}」, the precision={precision*100 :.2f}%, the recall={recall*100:.2f}%, f1={f1*100:.2f}%")
            precision = _care_div(all_tp,all_tp+all_fp)
            recall = _care_div(all_tp, all_tp+all_fn)
            f1 = _care_div(2*precision*recall, precision+recall)
            return precision, recall, f1

        all_predicts = []
        all_ground_truths = []
        for predict in predicts:
            text = predict['sentence']
            predict_ins = _group_entities_info(predict['predict_entities_info'], text)
            ground_truth_ins = _group_entities_info(predict['entities_info'], text)
            all_predicts.append(predict_ins)
            all_ground_truths.append(ground_truth_ins)

        precision, recall, f1 = _calc_score(all_predicts, all_ground_truths)
        real_name = self.loss_name_map(stage)
        logger.info(f'{real_name}_precision: {precision*100}, {real_name}_recall: {recall*100}, {real_name}_f1: {f1*100}')
        return {f'{real_name}_precision': precision*100, f'{real_name}_recall': recall*100, f'{real_name}_f1': f1*100}

    def do_save(self, predicts: List, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict, save_condition: bool=False):
        """save the predict when save_condition==True

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            save_condition: True for save, False for depend on rt_config

        Returns: 
            None

        """
        if self.config.start_save_epoch == -1 or self.config.start_save_step == -1:
            self.config.start_save_step = rt_config.get('total_steps', 0) - 1
            self.config.start_save_epoch = rt_config.get('total_epochs', 0) - 1
        if rt_config['current_step']>=self.config.start_save_step or rt_config['current_epoch']>=self.config.start_save_epoch:
            save_condition = True
        if save_condition:
            save_path = os.path.join(self.config.save_root_path, self.config.save_path.get(stage, ''))
            if "current_step" in rt_config:
                save_file = os.path.join(save_path, f"step_{str(rt_config['current_step'])}_predict.json")
            else:
                save_file = os.path.join(save_path, 'predict.json')
            logger.info(f"Save the {stage} predict data at {save_file}")
            with open(save_file, 'w') as f:
                json.dump(predicts, f, indent=4, ensure_ascii=False)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
from . import embedding_register, embedding_config_register
from typing import Dict, List, Set
from dlk.core.base_module import SimpleModule, BaseModuleConfig
from dlk.utils.io import open
import pickle as pkl
from dlk.utils.logger import Logger
import torch

logger = Logger.get_logger()

@embedding_config_register('static')
class StaticEmbeddingConfig(BaseModuleConfig):
    """Config for StaticEmbedding

    Config Example:
        >>> {
        >>>     "config": {
        >>>         "embedding_file": "*@*", //the embedding file, must be saved as numpy array by pickle
        >>>         "embedding_dim": "*@*",
        >>>         //if the embedding_file is a dict, you should provide the dict trace to embedding
        >>>         "embedding_trace": ".", //default the file itself is the embedding
        >>>         /*embedding_trace: "embedding", //this means the <embedding = pickle.load(embedding_file)["embedding"]>*/
        >>>         /*embedding_trace: "meta.embedding", //this means the <embedding = pickle.load(embedding_file)['meta']["embedding"]>*/
        >>>         "freeze": false, // is freeze
        >>>         "padding_idx": 0, //dropout rate
        >>>         "dropout": 0, //dropout rate
        >>>         "output_map": {},
        >>>         "input_map": {}, // required_key: provide_key
        >>>     },
        >>>     "_name": "static",
        >>> }
    """
    def __init__(self, config: Dict):
        super(StaticEmbeddingConfig, self).__init__(config)
        config = config['config']

        embedding_file = config['embedding_file']
        with open(embedding_file, 'rb') as f:
            embedding_file = pkl.load(f)
        embedding_trace = config["embedding_trace"]
        if embedding_trace != '.':
            traces = embedding_trace.split('.')
            for trace in traces:
                embedding_file = embedding_file[trace]
        self.embedding = embedding_file
        self.freeze = config['freeze']
        self.dropout = config['dropout']
        self.padding_idx = config['padding_idx']
        self.embedding_dim = config['embedding_dim']
        self.post_check(config, used=[
            "embedding_file",
            "embedding_dim",
            "embedding_trace",
            "freeze",
            "padding_idx",
            "dropout",
            "return_logits",
        ])


@embedding_register('static')
class StaticEmbedding(SimpleModule):
    """ from 'input_ids' generate static 'embedding' like glove, word2vec
    """

    def __init__(self, config: StaticEmbeddingConfig):
        super().__init__(config)
        self._provided_keys = set() # provided by privous module, will update by the check_keys_are_provided
        self._provide_keys = {'embedding'} # provide by this module
        self._required_keys = {'input_ids'} # required by this module
        self.config = config
        self.dropout = nn.Dropout(float(self.config.dropout))
        self.embedding = nn.Embedding.from_pretrained(torch.tensor(self.config.embedding, dtype=torch.float), freeze=self.config.freeze, padding_idx=self.config.padding_idx)
        assert self.embedding.weight.shape[-1] == self.config.embedding_dim

    def init_weight(self, method):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns: 
            None

        """
        logger.info(f'The static embedding is loaded the pretrained.')

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """get the pretrained static embedding like glove word2vec

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        inputs[self.get_output_name('embedding')] = self.dropout(self.embedding(inputs[self.get_input_name('input_ids')]))
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([inputs[self.get_output_name('embedding')]]))

        return inputs



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hjson
import os
from typing import Dict, Union, Callable, List, Any
from dlk.utils.parser import BaseConfigParser
from dlk.utils.config import ConfigTool
from dlk.data.datamodules import datamodule_register, datamodule_config_register
from dlk.managers import manager_register, manager_config_register
from dlk.core.imodels import imodel_register, imodel_config_register
import pickle as pkl
from dlk.utils.io import open
import json
from dlk.utils.logger import Logger

logger = Logger.get_logger()


class Train(object):
    """Trainer

    Config Example:
        >>> {
        >>>     "_focus": {
        >>>     },
        >>>     "_link": {},
        >>>     "_search": {},
        >>>     "config": {
        >>>         "save_dir": "*@*",  # must be provided
        >>>         "data_path": "*@*",  # must be provided
        >>>     },
        >>>     "task": {
        >>>         "_name": task_name
        >>>         ...
        >>>     }
        >>> }
    """
    def __init__(self, config: Union[str, Dict], ckpt: str = ""):
        super(Train, self).__init__()
        if not isinstance(config, dict):
            with open(config, 'r') as f:
                config = hjson.load(f, object_pairs_hook=dict)

        self.ckpt = ckpt
        self.focus = config.pop('_focus', {})
        self.configs = BaseConfigParser(config).parser_with_check()
        if self.ckpt:
            assert len(
                self.configs
            ) == 1, f"Reuse the checkpoint(ckpt is not none), you must provide the (only one) config which generate the checkpoint."

        self.config_names = []
        for possible_config in self.configs:
            config_name = []
            for source, to in self.focus.items():
                config_point = possible_config
                trace = source.split('.')
                for t in trace:
                    config_point = config_point[t]
                config_name.append(to + str(config_point))
            if config_name:
                self.config_names.append('_'.join(config_name))
            else:
                self.config_names.append(possible_config['root']['_name'])

        if len(self.config_names) != len(set(self.config_names)):
            for config, name in zip(self.configs, self.config_names):
                logger.info(
                    f"{name}:\n{json.dumps(config, indent=4, ensure_ascii=False)}"
                )
            raise NameError('The config_names is not unique.')

    def run(self):
        """run for all configs

        Returns: 
            None

        """
        logger.info(
            f"You have {len(self.config_names)} training config(s), they all will be run."
        )
        for i, (config, name) in enumerate(zip(self.configs,
                                               self.config_names)):
            logger.info(f"Runing the {i}th {name}...")
            self.run_oneturn(config, name)

    def dump_config(self, config: Dict, name: str):
        """dump the config and change the log file path to config['config']['save_dir']+name

        Args:
            config: {"config": {"save_dir": '..'}}
            name: config name

        Returns: 
            None

        """
        log_path = os.path.join(config.get('config').get('save_dir'), name)
        with open(os.path.join(config.get('config').get('save_dir'), name, "config.json"), 'w') as f:
            json.dump(
                {
                    "root": config,
                    "_focus": self.focus
                },
                f,
                ensure_ascii=False,
                indent=4
            )
        Logger.init_file_logger("log.txt", log_path)

    def run_oneturn(self, config, name):
        """run this config

        Args:
            config: {"root": '...'}
            name: config name

        Returns: 
            None

        """

        config = config['root']
        # save configure
        self.dump_config(config, name)

        # get data
        data = self.get_data(config)

        # set training manager
        manager = self.get_manager(config, name)

        # register devices for valid repeat
        # set datamodule
        datamodule = self.get_datamodule(config, data, world_size = manager.manager.world_size)

        # init imodel and inject the origin test and valid data
        imodel = self.get_imodel(config, data)

        # start training
        manager.fit(model=imodel, datamodule=datamodule)
        manager.test(model=imodel, datamodule=datamodule)

    def get_data(self, config):
        """get the data decided by config

        Args:
            config: {"config": {"data_path": '..'}}

        Returns: 
            loaded data

        """
        with open(config['config']['data_path'], 'rb') as f:
            self.data = pkl.load(f).get('data', {})
        return self.data

    def get_datamodule(self, config, data, world_size):
        """get the datamodule decided by config, and fit the data to datamodule

        Args:
            config: {"task": {"datamodule": '..'}}
            data: {"train": '..', 'valid': '..', ..}
            devices: when the devices >1 and repeat_for_valid is True, we will repeat the 

        Returns: 
            datamodule

        """
        DataModule, data_module_config = ConfigTool.get_leaf_module(
            datamodule_register, datamodule_config_register, 'datamodule',
            config['task']['datamodule'])
        data_module_config.world_size = world_size
        datamodule = DataModule(data_module_config, data)
        return datamodule

    def get_manager(self, config, name):
        """get the tranin/predict manager decided by config

        Args:
            config: {"task": {"manager": '..'}, "config": {"save_dir"}}
            name: the predict progress name

        Returns: 
            manager

        """
        Manager, manager_config = ConfigTool.get_leaf_module(
            manager_register, manager_config_register, 'manager',
            config.get('task').get('manager'))
        manager = Manager(manager_config,
                          rt_config={
                              "save_dir": config.get('config').get("save_dir"),
                              "name": name
                          })
        return manager

    def get_imodel(self, config, data):
        """get the imodel decided by config, and inject the origin test and valid data

        Args:
            config: {"task": {"imodel": '..'}}
            data: {"train": '..', 'valid': '..', ..}

        Returns: 
            imodel

        """
        IModel, imodel_config = ConfigTool.get_leaf_module(
            imodel_register, imodel_config_register, 'imodel',
            config.get('task').get('imodel'))
        imodel = IModel(imodel_config)
        if self.ckpt:
            logger.info(f"reuse the checkpoint at {self.ckpt}")
            imodel.load_from_checkpoint(self.ckpt)
        if 'valid' in data:
            imodel._origin_valid_data = data['valid']

        if 'test' in data:
            imodel._origin_test_data = data['test']

        return imodel



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
from . import encoder_register, encoder_config_register
from typing import Dict, List, Set
from dlk.core.base_module import SimpleModule, BaseModuleConfig
import torch


@encoder_config_register('identity')
class IdentityEncoderConfig(BaseModuleConfig):
    """Config for IdentityEncoder

    Config Example:
        >>> {
        >>>     "config": {
        >>>     },
        >>>     "_name": "identity",
        >>> }
    """
    def __init__(self, config):
        super(IdentityEncoderConfig, self).__init__(config)
        self.post_check(config['config'])

@encoder_register('identity')
class IdentityEncoder(SimpleModule):
    """Do nothing
    """

    def __init__(self, config: IdentityEncoderConfig):
        super().__init__(config)
        self.config = config

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """return inputs

        Args:
            inputs: anything

        Returns: 
            inputs 

        """
        return inputs
